# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mesop as me

from copycat.ui import components
from copycat.ui import event_handlers
from copycat.ui import states


def setup():
  """Renders the setup page.

  The setup page is used to configure some global Copycat parameters, such as
  the Google Cloud Project and Google Ads Account.
  """
  params = me.state(states.CopycatParamsState)
  with components.column(width="100%"):
    with components.rounded_box_section("Google Cloud Project"):
      with components.row():
        me.text(
            "Copycat uses Vertex AI on Google Cloud to call Gemini. You must "
            + "provide a valid Google Cloud Project for this, with "
            + "the Vertex AI API enabled.",
            style=me.Style(margin=me.Margin(bottom=15)),
        )

      with components.row():
        me.input(
            label="Google Cloud Project ID",
            key="vertex_ai_project_id",
            on_blur=event_handlers.update_copycat_parameter,
            value=params.vertex_ai_project_id,
            appearance="outline",
            style=me.Style(
                margin=me.Margin.symmetric(horizontal=5),
            ),
        )
        me.input(
            label="Google Cloud Project Location",
            key="vertex_ai_location",
            on_blur=event_handlers.update_copycat_parameter,
            value=params.vertex_ai_location,
            appearance="outline",
            style=me.Style(margin=me.Margin.symmetric(horizontal=5)),
        )



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for mock_gspread.

These test that the mock behaves like the real gspread client.
"""

from absl.testing import absltest
from absl.testing import parameterized
import gspread

from copycat.data import mock_gspread


class PatchEmbeddingsModelTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.gspread_patcher = mock_gspread.PatchGspread()
    self.gspread_patcher.start()

  def tearDown(self):
    super().tearDown()
    self.gspread_patcher.stop()

  def test_create_makes_an_empty_sheet(self):
    client = gspread.authorize(None)
    spreadsheet = client.create("test_spreadsheet")

    self.assertListEqual(
        [sheet.title for sheet in spreadsheet.worksheets()], ["Sheet1"]
    )
    self.assertListEqual(spreadsheet.worksheet("Sheet1").get_all_records(), [])

  def test_get_url_loads_the_sheet_from_the_url(self):
    client = gspread.authorize(None)
    spreadsheet = client.create("test_spreadsheet")

    retrieved_spreadsheet = client.open_by_url(spreadsheet.url)

    self.assertIs(retrieved_spreadsheet, spreadsheet)

  def test_get_url_raises_exception_if_url_is_not_found(self):
    client = gspread.authorize(None)
    spreadsheet = client.create("test_spreadsheet")

    with self.assertRaises(gspread.SpreadsheetNotFound):
      client.open_by_url("invalid url")

  def test_add_worksheet_adds_an_empty_worksheet_with_default_rows_and_columns(
      self,
  ):
    client = gspread.authorize(None)
    spreadsheet = client.create("test_spreadsheet")

    spreadsheet.add_worksheet("new worksheet")

    worksheet = spreadsheet.worksheet("new worksheet")
    self.assertListEqual(worksheet.get_all_records(), [])
    self.assertEqual(worksheet.row_count, 1000)
    self.assertEqual(worksheet.col_count, 26)

  def test_add_worksheet_adds_an_empty_worksheet_with_specified_rows_and_columns(
      self,
  ):
    client = gspread.authorize(None)
    spreadsheet = client.create("test_spreadsheet")

    spreadsheet.add_worksheet("new worksheet", rows=10, cols=2)

    worksheet = spreadsheet.worksheet("new worksheet")
    self.assertListEqual(worksheet.get_all_records(), [])
    self.assertEqual(worksheet.row_count, 10)
    self.assertEqual(worksheet.col_count, 2)

  def test_del_worksheet_deletes_the_worksheet(
      self,
  ):
    client = gspread.authorize(None)
    spreadsheet = client.create("test_spreadsheet")
    spreadsheet.add_worksheet("new worksheet")

    self.assertListEqual(
        [sheet.title for sheet in spreadsheet.worksheets()],
        ["Sheet1", "new worksheet"],
    )

    spreadsheet.del_worksheet(spreadsheet.worksheet("new worksheet"))
    self.assertListEqual(
        [sheet.title for sheet in spreadsheet.worksheets()], ["Sheet1"]
    )

  def test_worksheets_returns_all_worksheets_in_the_spreadsheet(self):
    client = gspread.authorize(None)
    spreadsheet = client.create("test_spreadsheet")
    spreadsheet.add_worksheet("new worksheet 1")
    spreadsheet.add_worksheet("new worksheet 2")

    self.assertListEqual(
        [worksheet.title for worksheet in spreadsheet.worksheets()],
        ["Sheet1", "new worksheet 1", "new worksheet 2"],
    )

  def test_get_all_records_returns_expected_records(self):
    client = gspread.authorize(None)
    spreadsheet = client.create("test_spreadsheet")
    worksheet = spreadsheet.worksheet("Sheet1")

    worksheet.update([["a", "b"], [1, 2], [3, 4]])

    self.assertListEqual(
        worksheet.get_all_records(),
        [{"a": 1, "b": 2}, {"a": 3, "b": 4}],
    )

  @parameterized.named_parameters([
      {
          "testcase_name": "freeze_rows",
          "rows": 1,
          "cols": None,
          "expected_frozen_rows": 1,
          "expected_frozen_cols": 0,
      },
      {
          "testcase_name": "freeze_cols",
          "rows": None,
          "cols": 2,
          "expected_frozen_rows": 0,
          "expected_frozen_cols": 2,
      },
      {
          "testcase_name": "freeze_both",
          "rows": 1,
          "cols": 2,
          "expected_frozen_rows": 1,
          "expected_frozen_cols": 2,
      },
  ])
  def test_can_freeze_rows_and_columns(
      self, rows, cols, expected_frozen_rows, expected_frozen_cols
  ):
    client = gspread.authorize(None)
    spreadsheet = client.create("test_spreadsheet")
    worksheet = spreadsheet.worksheet("Sheet1")

    worksheet.freeze(rows=rows, cols=cols)

    self.assertEqual(worksheet.frozen_row_count, expected_frozen_rows)
    self.assertEqual(worksheet.frozen_col_count, expected_frozen_cols)

  @parameterized.named_parameters([
      {
          "testcase_name": "single_range",
          "cell_range": "A1:B2",
          "expected_cells_with_formatting": [(0, 0), (0, 1), (1, 0), (1, 1)],
      },
      {
          "testcase_name": "single_cell",
          "cell_range": "A1",
          "expected_cells_with_formatting": [(0, 0)],
      },
      {
          "testcase_name": "complete_column",
          "cell_range": "A:A",
          "expected_cells_with_formatting": [(0, 0), (1, 0), (2, 0)],
      },
      {
          "testcase_name": "complete_row",
          "cell_range": "1:1",
          "expected_cells_with_formatting": [(0, 0), (0, 1), (0, 2)],
      },
  ])
  def test_format_adds_formatting_to_expected_cells(
      self, cell_range, expected_cells_with_formatting
  ):
    client = gspread.authorize(None)
    spreadsheet = client.create("test_spreadsheet")
    worksheet = spreadsheet.add_worksheet("Sheet2", rows=3, cols=3)

    target_format = {"textFormat": {"bold": True}}
    worksheet.format(cell_range, target_format)

    for i in range(worksheet.row_count):
      for j in range(worksheet.col_count):
        if (i, j) in expected_cells_with_formatting:
          self.assertDictEqual(worksheet._formatting[i][j], target_format)
        else:
          self.assertDictEqual(worksheet._formatting[i][j], {})

  @parameterized.named_parameters([
      {
          "testcase_name": "single_range",
          "cell_range": "B2:C3",
          "values": [[1, 2], [3, 4]],
          "expected_data": [["", "", ""], ["", 1, 2], ["", 3, 4]],
      },
      {
          "testcase_name": "single_cell",
          "cell_range": "A1",
          "values": [[1]],
          "expected_data": [[1, "", ""], ["", "", ""], ["", "", ""]],
      },
      {
          "testcase_name": "no range specified",
          "cell_range": None,
          "values": [[1, 2], [3, 4]],
          "expected_data": [[1, 2, ""], [3, 4, ""], ["", "", ""]],
      },
  ])
  def test_update_can_update_specific_cell_range(
      self, cell_range, values, expected_data
  ):
    client = gspread.authorize(None)
    spreadsheet = client.create("test_spreadsheet")
    worksheet = spreadsheet.add_worksheet("Sheet2", rows=3, cols=3)

    worksheet.update(values, cell_range)

    self.assertListEqual(worksheet._data, expected_data)

  @parameterized.named_parameters([
      {
          "testcase_name": "too many cols for range",
          "cell_range": "A1:A2",
          "values": [[1, 2], [3, 4]],
      },
      {
          "testcase_name": "too many rows for range",
          "cell_range": "A1:B1",
          "values": [[1, 2], [3, 4]],
      },
      {
          "testcase_name": "too many cols for sheet",
          "cell_range": "A1:D1",
          "values": [[1, 2, 3, 4]],
      },
      {
          "testcase_name": "too many rows for sheet",
          "cell_range": "A1:A4",
          "values": [[1], [2], [3], [4]],
      },
      {
          "testcase_name": "too many cols for sheet no range specified",
          "cell_range": None,
          "values": [[1, 2, 3, 4]],
      },
      {
          "testcase_name": "too many rows for sheet no range specified",
          "cell_range": None,
          "values": [[1], [2], [3], [4]],
      },
  ])
  def test_update_raises_value_error_if_out_of_range(self, cell_range, values):
    client = gspread.authorize(None)
    spreadsheet = client.create("test_spreadsheet")
    worksheet = spreadsheet.add_worksheet("Sheet2", rows=3, cols=3)

    with self.assertRaises(ValueError):
      worksheet.update(values, cell_range)

  def test_batch_update_applies_all_updates(self):
    client = gspread.authorize(None)
    spreadsheet = client.create("test_spreadsheet")
    worksheet = spreadsheet.add_worksheet("Sheet2", rows=4, cols=4)

    updates_list = [
        {"range": "A1:A2", "values": [["1"], ["2"]]},
        {"range": "B3:C4", "values": [["3", "4"], ["5", "6"]]},
    ]

    worksheet.batch_update(updates_list)

    expected_data = [
        ["1", "", "", ""],
        ["2", "", "", ""],
        ["", "3", "4", ""],
        ["", "5", "6", ""],
    ]
    self.assertListEqual(worksheet._data, expected_data)

  def test_add_rows_adds_rows_to_the_worksheet(self):
    client = gspread.authorize(None)
    spreadsheet = client.create("test_spreadsheet")
    worksheet = spreadsheet.add_worksheet("Sheet2", rows=4, cols=4)

    worksheet.add_rows(2)

    self.assertEqual(worksheet.row_count, 6)

  def test_add_cols_adds_cols_to_the_worksheet(self):
    client = gspread.authorize(None)
    spreadsheet = client.create("test_spreadsheet")
    worksheet = spreadsheet.add_worksheet("Sheet2", rows=4, cols=4)

    worksheet.add_cols(2)

    self.assertEqual(worksheet.col_count, 6)

  @parameterized.named_parameters([
      {
          "testcase_name": "no_end_index",
          "start_index": 1,
          "end_index": None,
          "expected_data": [
              ["1", "2", "", ""],
              ["5", "6", "", ""],
              ["7", "8", "", ""],
          ],
      },
      {
          "testcase_name": "with_end_index",
          "start_index": 1,
          "end_index": 2,
          "expected_data": [
              ["1", "2", "", ""],
              ["7", "8", "", ""],
          ],
      },
  ])
  def test_delete_rows_deletes_rows_from_the_worksheet(
      self, start_index, end_index, expected_data
  ):
    client = gspread.authorize(None)
    spreadsheet = client.create("test_spreadsheet")
    worksheet = spreadsheet.add_worksheet("Sheet2", rows=4, cols=4)
    worksheet.update([
        ["1", "2"],
        ["3", "4"],
        ["5", "6"],
        ["7", "8"],
    ])

    worksheet.delete_rows(start_index=start_index, end_index=end_index)

    self.assertListEqual(worksheet._data, expected_data)

  @parameterized.named_parameters([
      {
          "testcase_name": "no_end_index",
          "start_index": 1,
          "end_index": None,
          "expected_data": [
              ["1", "3", "4"],
              ["5", "7", "8"],
              ["", "", ""],
              ["", "", ""],
          ],
      },
      {
          "testcase_name": "with_end_index",
          "start_index": 1,
          "end_index": 2,
          "expected_data": [
              ["1", "4"],
              ["5", "8"],
              ["", ""],
              ["", ""],
          ],
      },
  ])
  def test_delete_columns_deletes_columns_from_the_worksheet(
      self, start_index, end_index, expected_data
  ):
    client = gspread.authorize(None)
    spreadsheet = client.create("test_spreadsheet")
    worksheet = spreadsheet.add_worksheet("Sheet2", rows=4, cols=4)
    worksheet.update([
        ["1", "2", "3", "4"],
        ["5", "6", "7", "8"],
    ])

    worksheet.delete_columns(start_index=start_index, end_index=end_index)

    self.assertListEqual(worksheet._data, expected_data)

  def test_clear_removes_all_data_and_formatting_from_the_worksheet(self):
    client = gspread.authorize(None)
    spreadsheet = client.create("test_spreadsheet")
    worksheet = spreadsheet.add_worksheet("Sheet2", rows=4, cols=4)
    worksheet.update([
        ["1", "2", "3", "4"],
        ["5", "6", "7", "8"],
    ])
    worksheet.format("A1:A2", {"textFormat": {"bold": True}})

    worksheet.clear()

    self.assertListEqual(worksheet._data, [[""] * 4] * 4)
    self.assertListEqual(worksheet._formatting, [[{}] * 4] * 4)

  def test_insert_row_inserts_a_row_at_the_given_index(self):
    client = gspread.authorize(None)
    spreadsheet = client.create("test_spreadsheet")
    worksheet = spreadsheet.add_worksheet("Sheet2", rows=2, cols=4)
    worksheet.update([
        ["1", "2", "3", "4"],
        ["5", "6", "7", "8"],
    ])

    worksheet.insert_row(["9", "10", "11", "12"], index=2)

    self.assertListEqual(
        worksheet._data,
        [
            ["1", "2", "3", "4"],
            ["9", "10", "11", "12"],
            ["5", "6", "7", "8"],
        ],
    )
    self.assertListEqual(
        worksheet._formatting,
        [
            [{}] * 4,
            [{}] * 4,
            [{}] * 4,
        ],
    )

  def test_insert_row_adds_columns_if_needed(self):
    client = gspread.authorize(None)
    spreadsheet = client.create("test_spreadsheet")
    worksheet = spreadsheet.add_worksheet("Sheet2", rows=2, cols=4)
    worksheet.update([
        ["1", "2", "3", "4"],
        ["5", "6", "7", "8"],
    ])

    worksheet.insert_row(["9", "10", "11", "12", "13"], index=2)

    self.assertEqual(worksheet.col_count, 5)
    self.assertListEqual(
        worksheet._data,
        [
            ["1", "2", "3", "4", ""],
            ["9", "10", "11", "12", "13"],
            ["5", "6", "7", "8", ""],
        ],
    )
    self.assertListEqual(
        worksheet._formatting,
        [
            [{}] * 5,
            [{}] * 5,
            [{}] * 5,
        ],
    )

  def test_row_values_returns_the_values_in_the_given_row(self):
    client = gspread.authorize(None)
    spreadsheet = client.create("test_spreadsheet")
    worksheet = spreadsheet.add_worksheet("Sheet2", rows=2, cols=4)
    worksheet.update([
        ["1", "2", "3", "4"],
        ["5", "6", "7", "8"],
    ])

    self.assertListEqual(worksheet.row_values(1), ["1", "2", "3", "4"])
    self.assertListEqual(worksheet.row_values(2), ["5", "6", "7", "8"])


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the sheets module."""

import datetime as dt
import logging

from absl.testing import absltest
from absl.testing import parameterized
import gspread
import pandas as pd

from copycat.data import mock_gspread
from copycat.data import sheets


class GoogleSheetTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.gspread_patcher = mock_gspread.PatchGspread()
    self.gspread_patcher.start()

    dummy_credentials = "dummy_credentials"
    sheets.set_google_auth_credentials(dummy_credentials)
    self.client = gspread.authorize(dummy_credentials)

  def tearDown(self):
    super().tearDown()
    self.gspread_patcher.stop()

  def test_can_instantiate_with_new_sheet(self):
    google_sheet = sheets.GoogleSheet.new("test_sheet")

    expected_str = "\n".join([
        "GoogleSheet:",
        "  URL: https://mock.sheets.com/spreadsheet/test_sheet",
        "  Name: test_sheet",
        "  Worksheet Names: ['Sheet1']",
    ])
    self.assertEqual(str(google_sheet), expected_str)

  def test_can_instantiate_with_existing_sheet(self):
    spreadsheet = self.client.create("test_sheet")
    spreadsheet.add_worksheet("Another Sheet")

    google_sheet = sheets.GoogleSheet.load(spreadsheet.url)

    expected_str = "\n".join([
        "GoogleSheet:",
        "  URL: https://mock.sheets.com/spreadsheet/test_sheet",
        "  Name: test_sheet",
        "  Worksheet Names: ['Sheet1', 'Another Sheet']",
    ])
    self.assertEqual(str(google_sheet), expected_str)

  def test_can_check_if_sheet_exists(self):
    google_sheet = sheets.GoogleSheet.new("test_sheet")

    self.assertIn("Sheet1", google_sheet)
    self.assertNotIn("Another Sheet", google_sheet)

  def test_can_load_data_as_pandas_dataframe(self):
    spreadsheet = self.client.create("test_sheet")
    spreadsheet.worksheet("Sheet1").update([
        ["header 1", "header 2"],
        ["row 1 col 1", "row 1 col 2"],
        ["row 2 col 1", "row 2 col 2"],
    ])

    google_sheet = sheets.GoogleSheet.load(spreadsheet.url)
    data = google_sheet["Sheet1"]

    expected_data = pd.DataFrame({
        "header 1": ["row 1 col 1", "row 2 col 1"],
        "header 2": ["row 1 col 2", "row 2 col 2"],
    })

    pd.testing.assert_frame_equal(data, expected_data)

  def test_can_load_data_with_index_as_pandas_dataframe(self):
    spreadsheet = self.client.create("test_sheet")
    spreadsheet.worksheet("Sheet1").update([
        ["my_index", "header 1", "header 2"],
        ["a", "row 1 col 1", "row 1 col 2"],
        ["b", "row 2 col 1", "row 2 col 2"],
    ])
    spreadsheet.worksheet("Sheet1").freeze(cols=1)  # Indexes are frozen columns

    google_sheet = sheets.GoogleSheet.load(spreadsheet.url)
    data = google_sheet["Sheet1"]

    expected_data = pd.DataFrame({
        "my_index": ["a", "b"],
        "header 1": ["row 1 col 1", "row 2 col 1"],
        "header 2": ["row 1 col 2", "row 2 col 2"],
    }).set_index("my_index")

    pd.testing.assert_frame_equal(data, expected_data)

  def test_loading_data_with_only_column_names_and_empty_row_returns_expected_data(
      self,
  ):
    spreadsheet = self.client.create("test_sheet")
    spreadsheet.add_worksheet("Sheet2", rows=2, cols=3)
    spreadsheet.worksheet("Sheet2").update(
        [["my_index", "header 1", "header 2"]]
    )
    spreadsheet.worksheet("Sheet2").freeze(cols=1)  # Indexes are frozen columns

    google_sheet = sheets.GoogleSheet.load(spreadsheet.url)
    data = google_sheet["Sheet2"]

    expected_data = pd.DataFrame(
        columns=[
            "my_index",
            "header 1",
            "header 2",
        ]
    ).set_index("my_index")

    pd.testing.assert_frame_equal(data, expected_data)

  def test_loading_data_with_only_column_names_returns_expected_data(self):
    spreadsheet = self.client.create("test_sheet")
    spreadsheet.add_worksheet("Sheet2", rows=1, cols=3)
    spreadsheet.worksheet("Sheet2").update(
        [["my_index", "header 1", "header 2"]]
    )
    spreadsheet.worksheet("Sheet2").freeze(cols=1)  # Indexes are frozen columns

    google_sheet = sheets.GoogleSheet.load(spreadsheet.url)
    data = google_sheet["Sheet2"]

    expected_data = pd.DataFrame(
        columns=[
            "my_index",
            "header 1",
            "header 2",
        ]
    ).set_index("my_index")

    pd.testing.assert_frame_equal(data, expected_data)

  def test_writing_data_writes_to_sheet(self):
    google_sheet = sheets.GoogleSheet.new("test_sheet")
    data = pd.DataFrame({
        "my_index": ["a", "b"],
        "header 1": ["row 1 col 1", "row 2 col 1"],
        "header 2": ["row 1 col 2", "row 2 col 2"],
    }).set_index("my_index")

    google_sheet["Sheet1"] = data

    spreadsheet = self.client.open_by_url(google_sheet.url)
    worksheet = spreadsheet.worksheet("Sheet1")
    self.assertListEqual(
        worksheet._data,
        [
            ["my_index", "header 1", "header 2"],
            ["a", "row 1 col 1", "row 1 col 2"],
            ["b", "row 2 col 1", "row 2 col 2"],
        ],
    )

  def test_writing_data_updates_size_of_worksheet_to_match_data(self):
    google_sheet = sheets.GoogleSheet.new("test_sheet")
    spreadsheet = self.client.open_by_url(google_sheet.url)
    worksheet = spreadsheet.worksheet("Sheet1")

    self.assertEqual(worksheet.row_count, 1000)
    self.assertEqual(worksheet.col_count, 26)

    google_sheet["Sheet1"] = pd.DataFrame({
        "my_index": ["a", "b"],
        "header 1": ["row 1 col 1", "row 2 col 1"],
        "header 2": ["row 1 col 2", "row 2 col 2"],
    }).set_index("my_index")

    self.assertEqual(worksheet.row_count, 3)
    self.assertEqual(worksheet.col_count, 3)

  def test_writing_sets_format_of_column_names(self):
    google_sheet = sheets.GoogleSheet.new("test_sheet")

    google_sheet["Sheet1"] = pd.DataFrame({
        "my_index": ["a", "b"],
        "header 1": ["row 1 col 1", "row 2 col 1"],
        "header 2": ["row 1 col 2", "row 2 col 2"],
    }).set_index("my_index")

    spreadsheet = self.client.open_by_url(google_sheet.url)
    worksheet = spreadsheet.worksheet("Sheet1")

    for cell_format in worksheet._formatting[0]:
      # First row should be formatted like the header
      self.assertDictEqual(cell_format, sheets.HEADING_FORMAT)

    for row_format in worksheet._formatting[1:]:
      for cell_format in row_format:
        # All other rows should not be formatted
        self.assertDictEqual(cell_format, {})

  def test_writing_freezes_the_index_and_columns(self):
    google_sheet = sheets.GoogleSheet.new("test_sheet")
    spreadsheet = self.client.open_by_url(google_sheet.url)
    worksheet = spreadsheet.worksheet("Sheet1")

    self.assertEqual(worksheet.frozen_row_count, 0)
    self.assertEqual(worksheet.frozen_col_count, 0)

    google_sheet["Sheet1"] = pd.DataFrame({
        "my_index": ["a", "b"],
        "header 1": ["row 1 col 1", "row 2 col 1"],
        "header 2": ["row 1 col 2", "row 2 col 2"],
    }).set_index("my_index")

    self.assertEqual(worksheet.frozen_row_count, 1)
    self.assertEqual(worksheet.frozen_col_count, 1)

  def test_writing_data_adds_worksheet_if_needed(self):
    google_sheet = sheets.GoogleSheet.new("test_sheet")
    spreadsheet = self.client.open_by_url(google_sheet.url)

    worksheet_names = [sheet.title for sheet in spreadsheet.worksheets()]
    self.assertListEqual(worksheet_names, ["Sheet1"])

    google_sheet["Sheet2"] = pd.DataFrame({
        "my_index": ["a", "b"],
        "header 1": ["row 1 col 1", "row 2 col 1"],
        "header 2": ["row 1 col 2", "row 2 col 2"],
    }).set_index("my_index")

    worksheet_names = [sheet.title for sheet in spreadsheet.worksheets()]
    self.assertListEqual(worksheet_names, ["Sheet1", "Sheet2"])

  def test_writing_changed_data_overwrites_existing_data(self):
    google_sheet = sheets.GoogleSheet.new("test_sheet")
    google_sheet["Sheet1"] = pd.DataFrame({
        "my_index": ["a", "b", "c", "d"],
        "header 1": ["1", "2", "3", "4"],
        "header 2": ["5", "6", "7", "8"],
    }).set_index("my_index")

    # Some rows changed, some rows added, some rows unchanged
    new_data = pd.DataFrame({
        "my_index": ["a", "b change", "c", "d change", "e"],
        "header 1": ["1 change", "2", "3", "4 change", "9"],
        "header 2": ["5", "6 change", "7", "8", "10"],
    }).set_index("my_index")
    google_sheet["Sheet1"] = new_data

    pd.testing.assert_frame_equal(google_sheet["Sheet1"], new_data)

  def test_writing_data_with_different_columns_overwrites_existing_data(self):
    google_sheet = sheets.GoogleSheet.new("test_sheet")
    google_sheet["Sheet1"] = pd.DataFrame({
        "my_index": ["a", "b", "c", "d"],
        "header 1": ["1", "2", "3", "4"],
        "header 2": ["5", "6", "7", "8"],
    }).set_index("my_index")

    # Column name changed
    new_data = pd.DataFrame({
        "my_index": ["a", "b", "c", "d"],
        "header 1 changed": ["1", "2", "3", "4"],
        "header 2": ["5", "6", "7", "8"],
    }).set_index("my_index")
    google_sheet["Sheet1"] = new_data

    pd.testing.assert_frame_equal(google_sheet["Sheet1"], new_data)

  def test_delete_worksheet_deletes_the_worksheet(self):
    google_sheet = sheets.GoogleSheet.new("test_sheet")
    google_sheet["Sheet2"] = pd.DataFrame({
        "my_index": ["a", "b", "c", "d"],
        "header 1": ["1", "2", "3", "4"],
        "header 2": ["5", "6", "7", "8"],
    }).set_index("my_index")

    self.assertIn("Sheet2", google_sheet)
    google_sheet.delete_worksheet("Sheet2")
    self.assertNotIn("Sheet2", google_sheet)


class GoogleSheetLoggerTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.gspread_patcher = mock_gspread.PatchGspread()
    self.gspread_patcher.start()

    dummy_credentials = "dummy_credentials"
    sheets.set_google_auth_credentials(dummy_credentials)
    self.client = gspread.authorize(dummy_credentials)

  def tearDown(self):
    super().tearDown()
    self.gspread_patcher.stop()

  def test_can_write_logs_if_logs_tab_does_not_exist(self):
    spreadsheet = self.client.create("test_spreadsheet")

    logger = logging.getLogger("test_logger")
    handler = sheets.GoogleSheetsHandler(spreadsheet.url)
    handler.setLevel(logging.INFO)
    logger.handlers = []
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)

    logger.info("test info message")
    logger.warning("test warning message")
    logger.error("test error message")
    logger.critical("test critical message")

    worksheet = spreadsheet.worksheet("Logs")
    data_without_timestamp_col = [row[1:] for row in worksheet._data]
    timestamp_col = [row[0] for row in worksheet._data]

    self.assertListEqual(
        data_without_timestamp_col,
        [
            ["Log Level", "Logger Name", "Message"],
            ["CRITICAL", "test_logger", "test critical message"],
            ["ERROR", "test_logger", "test error message"],
            ["WARNING", "test_logger", "test warning message"],
            ["INFO", "test_logger", "test info message"],
            ["", "", ""],
        ],
    )
    self.assertEqual(timestamp_col[0], "UTC Timestamp")
    for timestamp in timestamp_col[1:-1]:
      dt.datetime.strptime(
          timestamp, "%Y-%m-%d %H:%M:%S"
      )  # Will raise if invalid format

  def test_adds_logs_if_logs_tab_exists(self):
    spreadsheet = self.client.create("test_spreadsheet")
    spreadsheet.add_worksheet("Logs", rows=3, cols=4)
    spreadsheet.worksheet("Logs").update(
        values=[
            sheets.GoogleSheetsLogSender.HEADINGS,
            [
                "2024-01-01 00:00:00",
                "INFO",
                "test_logger",
                "existing test info message",
            ],
        ],
    )

    logger = logging.getLogger("test_logger")
    handler = sheets.GoogleSheetsHandler(spreadsheet.url)
    handler.setLevel(logging.INFO)
    logger.handlers = []
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)

    logger.info("new info message")

    worksheet = spreadsheet.worksheet("Logs")
    data_without_timestamp_col = [row[1:] for row in worksheet._data]

    self.assertListEqual(
        data_without_timestamp_col,
        [
            ["Log Level", "Logger Name", "Message"],
            ["INFO", "test_logger", "new info message"],
            ["INFO", "test_logger", "existing test info message"],
            ["", "", ""],
        ],
    )

  def test_raises_error_if_logs_tab_has_wrong_headings(self):
    spreadsheet = self.client.create("test_spreadsheet")
    spreadsheet.add_worksheet("Logs", rows=2, cols=4)
    spreadsheet.worksheet("Logs").update(
        values=[["Wrong", "Headings"]],
    )

    with self.assertRaises(ValueError):
      sheets.GoogleSheetsLogSender(spreadsheet.url)


class CreateTemplateCopycatSheetTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.gspread_patcher = mock_gspread.PatchGspread()
    self.gspread_patcher.start()

    dummy_credentials = "dummy_credentials"
    sheets.set_google_auth_credentials(dummy_credentials)
    self.client = gspread.authorize(dummy_credentials)

  def tearDown(self):
    super().tearDown()
    self.gspread_patcher.stop()

  def test_expected_number_of_rows_are_created_when_include_demo_data_is_true(
      self,
  ):
    url = sheets.create_template_copycat_sheet(include_demo_data=True)
    sheet = sheets.GoogleSheet.load(url)

    self.assertLen(sheet[sheets.TEMPLATE_EXISTING_ADS_WORKSHEET_NAME], 7)
    self.assertLen(sheet[sheets.TEMPLATE_NEW_KEYWORDS_WORKSHEET_NAME], 14)
    self.assertLen(sheet[sheets.TEMPLATE_EXTRA_INSTRUCTIONS_WORKSHEET_NAME], 1)

  def test_no_rows_are_created_when_include_demo_data_is_false(self):
    url = sheets.create_template_copycat_sheet(include_demo_data=False)
    sheet = sheets.GoogleSheet.load(url)

    self.assertEmpty(sheet[sheets.TEMPLATE_EXISTING_ADS_WORKSHEET_NAME])
    self.assertEmpty(sheet[sheets.TEMPLATE_NEW_KEYWORDS_WORKSHEET_NAME])
    self.assertEmpty(sheet[sheets.TEMPLATE_EXTRA_INSTRUCTIONS_WORKSHEET_NAME])

  @parameterized.parameters(True, False)
  def test_expected_columns_and_index_are_created(self, include_demo_data):
    url = sheets.create_template_copycat_sheet(
        include_demo_data=include_demo_data
    )
    sheet = sheets.GoogleSheet.load(url)

    self.assertListEqual(
        sheet[
            sheets.TEMPLATE_EXISTING_ADS_WORKSHEET_NAME
        ].columns.values.tolist(),
        [
            "URL",
            "Ad Strength",
            "Keywords",
            "Headline 1",
            "Headline 2",
            "Headline 3",
            "Headline 4",
            "Headline 5",
            "Headline 6",
            "Headline 7",
            "Headline 8",
            "Headline 9",
            "Headline 10",
            "Headline 11",
            "Headline 12",
            "Headline 13",
            "Headline 14",
            "Headline 15",
            "Description 1",
            "Description 2",
            "Description 3",
            "Description 4",
        ],
    )
    self.assertListEqual(
        sheet[
            sheets.TEMPLATE_NEW_KEYWORDS_WORKSHEET_NAME
        ].columns.values.tolist(),
        [
            "Keyword",
        ],
    )
    self.assertListEqual(
        sheet[
            sheets.TEMPLATE_EXTRA_INSTRUCTIONS_WORKSHEET_NAME
        ].columns.values.tolist(),
        [
            "Extra Instructions",
        ],
    )

    self.assertListEqual(
        sheet[sheets.TEMPLATE_EXISTING_ADS_WORKSHEET_NAME].index.names,
        ["Campaign ID", "Ad Group"],
    )
    self.assertListEqual(
        sheet[sheets.TEMPLATE_NEW_KEYWORDS_WORKSHEET_NAME].index.names,
        ["Campaign ID", "Ad Group"],
    )
    self.assertListEqual(
        sheet[sheets.TEMPLATE_EXTRA_INSTRUCTIONS_WORKSHEET_NAME].index.names,
        ["Campaign ID", "Ad Group", "Version"],
    )


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized
import pandas as pd

from copycat import copycat
from copycat import google_ads
from copycat.data import utils


class ExplodeAndCollapseHeadlinesAndDescriptionsTest(parameterized.TestCase):

  @parameterized.named_parameters(
      dict(
          testcase_name="with_headlines_and_descriptions",
          data=pd.DataFrame({
              "Headline 1": ["a", "b"],
              "Headline 2": ["c", "--"],
              "Headline 3": ["d", ""],
              "Description 1": ["e", "f"],
              "Description 2": ["g", "--"],
              "Other column": [1, 2],
          }),
          expected=pd.DataFrame({
              "headlines": [["a", "c", "d"], ["b"]],
              "descriptions": [["e", "g"], ["f"]],
              "Other column": [1, 2],
          }),
      ),
      dict(
          testcase_name="with_no_headlines",
          data=pd.DataFrame({
              "Description 1": ["e", "f"],
              "Description 2": ["g", "--"],
              "Other column": [1, 2],
          }),
          expected=pd.DataFrame({
              "headlines": [[], []],
              "descriptions": [["e", "g"], ["f"]],
              "Other column": [1, 2],
          }),
      ),
      dict(
          testcase_name="with_no_descriptions",
          data=pd.DataFrame({
              "Headline 1": ["a", "b"],
              "Headline 2": ["c", "--"],
              "Headline 3": ["d", ""],
              "Other column": [1, 2],
          }),
          expected=pd.DataFrame({
              "headlines": [["a", "c", "d"], ["b"]],
              "descriptions": [[], []],
              "Other column": [1, 2],
          }),
      ),
      dict(
          testcase_name="with_no_headlines_or_descriptions",
          data=pd.DataFrame({"Other column": [1, 2]}),
          expected=pd.DataFrame({
              "headlines": [[], []],
              "descriptions": [[], []],
              "Other column": [1, 2],
          }),
      ),
  )
  def test_collapse_headlines_and_descriptions(self, data, expected):
    actual = utils.collapse_headlines_and_descriptions(data)
    pd.testing.assert_frame_equal(actual, expected, check_like=True)

  @parameterized.named_parameters(
      dict(
          testcase_name="with_headlines_and_descriptions",
          data=pd.DataFrame({
              "headlines": [["a", "c", "d"], ["b"]],
              "descriptions": [["e", "g"], ["f"]],
              "Other column": [1, 2],
          }),
          expected=pd.DataFrame({
              "Headline 1": ["a", "b"],
              "Headline 2": ["c", "--"],
              "Headline 3": ["d", "--"],
              "Description 1": ["e", "f"],
              "Description 2": ["g", "--"],
              "Other column": [1, 2],
          }),
      ),
      dict(
          testcase_name="with_no_headlines",
          data=pd.DataFrame({
              "headlines": [[], []],
              "descriptions": [["e", "g"], ["f"]],
              "Other column": [1, 2],
          }),
          expected=pd.DataFrame({
              "Description 1": ["e", "f"],
              "Description 2": ["g", "--"],
              "Other column": [1, 2],
          }),
      ),
      dict(
          testcase_name="with_no_descriptions",
          data=pd.DataFrame({
              "headlines": [["a", "c", "d"], ["b"]],
              "descriptions": [[], []],
              "Other column": [1, 2],
          }),
          expected=pd.DataFrame({
              "Headline 1": ["a", "b"],
              "Headline 2": ["c", "--"],
              "Headline 3": ["d", "--"],
              "Other column": [1, 2],
          }),
      ),
      dict(
          testcase_name="with_no_headlines_or_descriptions",
          data=pd.DataFrame({
              "headlines": [[], []],
              "descriptions": [[], []],
              "Other column": [1, 2],
          }),
          expected=pd.DataFrame({"Other column": [1, 2]}),
      ),
  )
  def test_explode_headlines_and_descriptions(self, data, expected):
    actual = utils.explode_headlines_and_descriptions(data)
    pd.testing.assert_frame_equal(actual, expected, check_like=True)

  @parameterized.named_parameters(
      dict(
          testcase_name="with_headlines_and_descriptions",
          data=pd.DataFrame({
              "headlines": [["a", "c", "d"], ["b"]],
              "descriptions": [["e", "g"], ["f"]],
              "Other column": [1, 2],
          }),
          expected=pd.DataFrame({
              "Headline 1": ["a", "b"],
              "Headline 2": ["c", "--"],
              "Headline 3": ["d", "--"],
              "Headline 4": ["--", "--"],
              "Headline 5": ["--", "--"],
              "Description 1": ["e", "f"],
              "Description 2": ["g", "--"],
              "Description 3": ["--", "--"],
              "Other column": [1, 2],
          }),
      ),
      dict(
          testcase_name="with_no_headlines",
          data=pd.DataFrame({
              "headlines": [[], []],
              "descriptions": [["e", "g"], ["f"]],
              "Other column": [1, 2],
          }),
          expected=pd.DataFrame({
              "Headline 1": ["--", "--"],
              "Headline 2": ["--", "--"],
              "Headline 3": ["--", "--"],
              "Headline 4": ["--", "--"],
              "Headline 5": ["--", "--"],
              "Description 1": ["e", "f"],
              "Description 2": ["g", "--"],
              "Description 3": ["--", "--"],
              "Other column": [1, 2],
          }),
      ),
      dict(
          testcase_name="with_no_descriptions",
          data=pd.DataFrame({
              "headlines": [["a", "c", "d"], ["b"]],
              "descriptions": [[], []],
              "Other column": [1, 2],
          }),
          expected=pd.DataFrame({
              "Headline 1": ["a", "b"],
              "Headline 2": ["c", "--"],
              "Headline 3": ["d", "--"],
              "Headline 4": ["--", "--"],
              "Headline 5": ["--", "--"],
              "Description 1": ["--", "--"],
              "Description 2": ["--", "--"],
              "Description 3": ["--", "--"],
              "Other column": [1, 2],
          }),
      ),
      dict(
          testcase_name="with_no_headlines_or_descriptions",
          data=pd.DataFrame({
              "headlines": [[], []],
              "descriptions": [[], []],
              "Other column": [1, 2],
          }),
          expected=pd.DataFrame({
              "Headline 1": ["--", "--"],
              "Headline 2": ["--", "--"],
              "Headline 3": ["--", "--"],
              "Headline 4": ["--", "--"],
              "Headline 5": ["--", "--"],
              "Description 1": ["--", "--"],
              "Description 2": ["--", "--"],
              "Description 3": ["--", "--"],
              "Other column": [1, 2],
          }),
      ),
  )
  def test_explode_headlines_and_descriptions_applies_max_headlines_and_descriptions(
      self, data, expected
  ):
    actual = utils.explode_headlines_and_descriptions(
        data, max_headlines=5, max_descriptions=3
    )
    pd.testing.assert_frame_equal(actual, expected, check_like=True)

  @parameterized.named_parameters(
      dict(
          testcase_name="too many headlines",
          max_headlines=2,
          max_descriptions=3,
      ),
      dict(
          testcase_name="too many descriptions",
          max_headlines=5,
          max_descriptions=1,
      ),
  )
  def test_explode_headlines_and_descriptions_raises_value_error_if_max_headlines_or_descriptions_too_small(
      self, max_headlines, max_descriptions
  ):
    data = pd.DataFrame({
        "headlines": [["a", "c", "d"], ["b"]],
        "descriptions": [["e", "g"], ["f"]],
        "Other column": [1, 2],
    })
    with self.assertRaises(ValueError):
      utils.explode_headlines_and_descriptions(
          data, max_headlines=max_headlines, max_descriptions=max_descriptions
      )

  def test_explode_headlines_and_descriptions_raises_value_error_if_index_not_unique(
      self,
  ):
    data = pd.DataFrame({
        "headlines": [["a", "c", "d"], ["b"]],
        "descriptions": [["e", "g"], ["f"]],
        "Other column": [1, 1],
    }).set_index("Other column")
    with self.assertRaises(ValueError):
      utils.explode_headlines_and_descriptions(data)

  @parameterized.parameters(
      "headlines",
      "descriptions",
  )
  def test_explode_headlines_and_descriptions_raises_value_error_if_headlines_or_descriptions_are_not_lists(
      self, column_name
  ):
    data = pd.DataFrame({
        "headlines": [["a", "c", "d"], ["b"]],
        "descriptions": [["e", "g"], ["f"]],
        "Other column": [1, 1],
    })
    data[column_name] = ["a", "b"]  # Column does not contain lists.
    with self.assertRaises(ValueError):
      utils.explode_headlines_and_descriptions(data)


class IterateOverBatchesTest(parameterized.TestCase):

  def test_iterate_over_batches(self):
    data = pd.DataFrame({"a": [1, 2, 3, 4, 5, 6, 7]})
    batches = list(utils.iterate_over_batches(data, batch_size=3))

    self.assertLen(batches, 3)
    pd.testing.assert_frame_equal(batches[0], data.iloc[:3])
    pd.testing.assert_frame_equal(batches[1], data.iloc[3:6])
    pd.testing.assert_frame_equal(batches[2], data.iloc[6:7])

  def test_iterate_over_batches_with_limit_rows(self):
    data = pd.DataFrame({"a": [1, 2, 3, 4, 5, 6, 7]})
    batches = list(utils.iterate_over_batches(data, batch_size=3, limit_rows=5))

    self.assertLen(batches, 2)
    pd.testing.assert_frame_equal(batches[0], data.iloc[:3])
    pd.testing.assert_frame_equal(batches[1], data.iloc[3:5])

  def test_iterate_over_batches_with_too_large_limit_rows(self):
    data = pd.DataFrame({"a": [1, 2, 3, 4, 5, 6, 7]})
    batches = list(
        utils.iterate_over_batches(data, batch_size=3, limit_rows=20)
    )

    self.assertLen(batches, 3)
    pd.testing.assert_frame_equal(batches[0], data.iloc[:3])
    pd.testing.assert_frame_equal(batches[1], data.iloc[3:6])
    pd.testing.assert_frame_equal(batches[2], data.iloc[6:7])


class ConstructGenerationDataTest(parameterized.TestCase):

  def test_construct_generation_data_with_new_keywords_data_only(self):
    new_keywords_data = pd.DataFrame({
        "index_column_1": ["a", "a", "b", "b"],
        "index_column_2": ["c", "d", "d", "d"],
        "keyword": ["keyword 1", "keyword 2", "keyword 3", "keyword 4"],
    }).set_index(["index_column_1", "index_column_2"])

    actual = utils.construct_generation_data(
        new_keywords_data=new_keywords_data
    )

    expected = pd.DataFrame({
        "index_column_1": ["a", "a", "b"],
        "index_column_2": ["c", "d", "d"],
        "version": ["1", "1", "1"],
        "keywords": ["keyword 1", "keyword 2", "keyword 3, keyword 4"],
        "existing_headlines": [[], [], []],
        "existing_descriptions": [[], [], []],
        "additional_instructions": ["", "", ""],
    }).set_index(["index_column_1", "index_column_2", "version"])

    pd.testing.assert_frame_equal(actual, expected, check_like=True)

  def test_construct_generation_data_with_new_keywords_data_only_multiple_versions(
      self,
  ):
    new_keywords_data = pd.DataFrame({
        "index_column_1": ["a", "a", "b", "b"],
        "index_column_2": ["c", "d", "d", "d"],
        "keyword": ["keyword 1", "keyword 2", "keyword 3", "keyword 4"],
    }).set_index(["index_column_1", "index_column_2"])

    actual = utils.construct_generation_data(
        new_keywords_data=new_keywords_data,
        n_versions=2,
    )

    expected = pd.DataFrame({
        "index_column_1": ["a", "a", "a", "a", "b", "b"],
        "index_column_2": ["c", "c", "d", "d", "d", "d"],
        "version": ["1", "2", "1", "2", "1", "2"],
        "keywords": [
            "keyword 1",
            "keyword 1",
            "keyword 2",
            "keyword 2",
            "keyword 3, keyword 4",
            "keyword 3, keyword 4",
        ],
        "existing_headlines": [[], [], [], [], [], []],
        "existing_descriptions": [[], [], [], [], [], []],
        "additional_instructions": ["", "", "", "", "", ""],
    }).set_index(["index_column_1", "index_column_2", "version"])

    pd.testing.assert_frame_equal(actual, expected, check_like=True)

  def test_construct_generation_data_with_existing_generations_data(self):
    new_keywords_data = pd.DataFrame({
        "index_column_1": ["a", "a", "b", "b"],
        "index_column_2": ["c", "d", "d", "d"],
        "keyword": ["keyword 1", "keyword 2", "keyword 3", "keyword 4"],
    }).set_index(["index_column_1", "index_column_2"])

    existing_generations_data = pd.DataFrame({
        "index_column_1": ["a", "a"],
        "index_column_2": ["c", "d"],
        "version": ["1", "1"],
        "existing_headlines": [["headline 1"], ["headline 2"]],
        "existing_descriptions": [
            ["description 1"],
            ["description 2"],
        ],
    }).set_index(["index_column_1", "index_column_2", "version"])

    actual = utils.construct_generation_data(
        new_keywords_data=new_keywords_data,
        existing_generations_data=existing_generations_data,
    )

    expected = pd.DataFrame({
        "index_column_1": ["a", "a", "b"],
        "index_column_2": ["c", "d", "d"],
        "version": ["1", "1", "1"],
        "keywords": ["keyword 1", "keyword 2", "keyword 3, keyword 4"],
        "existing_headlines": [["headline 1"], ["headline 2"], []],
        "existing_descriptions": [
            ["description 1"],
            ["description 2"],
            [],
        ],
        "additional_instructions": ["", "", ""],
    }).set_index(["index_column_1", "index_column_2", "version"])

    pd.testing.assert_frame_equal(actual, expected, check_like=True)

  def test_construct_generation_data_raises_value_error_if_existing_generations_data_not_unique(
      self,
  ):
    new_keywords_data = pd.DataFrame({
        "index_column_1": ["a", "a", "b", "b"],
        "index_column_2": ["c", "d", "d", "d"],
        "keyword": ["keyword 1", "keyword 2", "keyword 3", "keyword 4"],
    }).set_index(["index_column_1", "index_column_2"])

    existing_generations_data = pd.DataFrame({
        "index_column_1": ["a", "a"],
        "index_column_2": ["c", "c"],
        "version": ["1", "1"],
        "existing_headlines": [["headline 1"], ["headline 2"]],
        "existing_descriptions": [
            ["description 1"],
            ["description 2"],
        ],
    }).set_index(["index_column_1", "index_column_2", "version"])

    with self.assertRaisesWithLiteralMatch(
        ValueError,
        "The index columns of the existing_generations_data are not unique,"
        " cannot merge with the new keywords data.",
    ):
      utils.construct_generation_data(
          new_keywords_data=new_keywords_data,
          existing_generations_data=existing_generations_data,
      )

  def test_construct_generation_data_with_additional_instructions(self):
    new_keywords_data = pd.DataFrame({
        "index_column_1": ["a", "a", "b", "b"],
        "index_column_2": ["c", "d", "d", "d"],
        "keyword": ["keyword 1", "keyword 2", "keyword 3", "keyword 4"],
    }).set_index(["index_column_1", "index_column_2"])

    additional_instructions_data = pd.DataFrame({
        "index_column_1": ["a", "a", "a", "a"],
        "index_column_2": ["c", "d", "c", "c"],
        "version": ["1", "1", "2", "2"],
        "additional_instructions": [
            "instruction 1",
            "instruction 2",
            "instruction 3",
            "instruction 4",
        ],
        "existing_headlines": [[], [], [], []],
        "existing_descriptions": [[], [], [], []],
    }).set_index(["index_column_1", "index_column_2", "version"])

    actual = utils.construct_generation_data(
        new_keywords_data=new_keywords_data,
        additional_instructions_data=additional_instructions_data,
        n_versions=2,
    )

    expected = pd.DataFrame({
        "index_column_1": ["a", "a", "a", "a", "b", "b"],
        "index_column_2": ["c", "c", "d", "d", "d", "d"],
        "version": ["1", "2", "1", "2", "1", "2"],
        "keywords": [
            "keyword 1",
            "keyword 1",
            "keyword 2",
            "keyword 2",
            "keyword 3, keyword 4",
            "keyword 3, keyword 4",
        ],
        "additional_instructions": [
            "instruction 1",
            "instruction 3\ninstruction 4",
            "instruction 2",
            "",
            "",
            "",
        ],
        "existing_headlines": [[], [], [], [], [], []],
        "existing_descriptions": [[], [], [], [], [], []],
    }).set_index(["index_column_1", "index_column_2", "version"])

    pd.testing.assert_frame_equal(actual, expected, check_like=True)

  def test_construct_generation_data_with_additional_instructions_with_all(
      self,
  ):
    new_keywords_data = pd.DataFrame({
        "index_column_1": ["a", "a", "b", "b"],
        "index_column_2": ["c", "d", "d", "d"],
        "keyword": ["keyword 1", "keyword 2", "keyword 3", "keyword 4"],
    }).set_index(["index_column_1", "index_column_2"])

    additional_instructions_data = pd.DataFrame({
        "index_column_1": [
            "a",
            "__ALL__",
            "b",
            "__ALL__",
            "a",
            "__ALL__",
            "b",
            "__ALL__",
        ],
        "index_column_2": [
            "c",
            "d",
            "__ALL__",
            "__ALL__",
            "c",
            "d",
            "__ALL__",
            "__ALL__",
        ],
        "version": [
            "1",
            "1",
            "1",
            "1",
            "__ALL__",
            "__ALL__",
            "__ALL__",
            "__ALL__",
        ],
        "additional_instructions": [
            "(a,c,1)",
            "(all,d,1)",
            "(b,all,1)",
            "(all,all,1)",
            "(a,c,all)",
            "(all,d,all)",
            "(b,all,all)",
            "(all,all,all)",
        ],
    }).set_index(["index_column_1", "index_column_2", "version"])

    actual = utils.construct_generation_data(
        new_keywords_data=new_keywords_data,
        additional_instructions_data=additional_instructions_data,
        n_versions=2,
    )

    expected = pd.DataFrame({
        "index_column_1": ["a", "a", "a", "a", "b", "b"],
        "index_column_2": ["c", "c", "d", "d", "d", "d"],
        "version": ["1", "2", "1", "2", "1", "2"],
        "keywords": [
            "keyword 1",
            "keyword 1",
            "keyword 2",
            "keyword 2",
            "keyword 3, keyword 4",
            "keyword 3, keyword 4",
        ],
        "additional_instructions": [
            "(a,c,1)\n(a,c,all)\n(all,all,1)\n(all,all,all)",
            "(a,c,all)\n(all,all,all)",
            "(all,all,1)\n(all,all,all)\n(all,d,1)\n(all,d,all)",
            "(all,all,all)\n(all,d,all)",
            "(all,all,1)\n(all,all,all)\n(all,d,1)\n(all,d,all)\n(b,all,1)\n(b,all,all)",
            "(all,all,all)\n(all,d,all)\n(b,all,all)",
        ],
        "existing_headlines": [[], [], [], [], [], []],
        "existing_descriptions": [[], [], [], [], [], []],
    }).set_index(["index_column_1", "index_column_2", "version"])

    pd.testing.assert_frame_equal(actual, expected, check_like=True)

  def test_construct_generation_data_with_empty_additional_instructions(
      self,
  ):
    new_keywords_data = pd.DataFrame({
        "index_column_1": ["a", "a", "b", "b"],
        "index_column_2": ["c", "d", "d", "d"],
        "keyword": ["keyword 1", "keyword 2", "keyword 3", "keyword 4"],
    }).set_index(["index_column_1", "index_column_2"])

    additional_instructions_data = pd.DataFrame(
        columns=[
            "index_column_1",
            "index_column_2",
            "version",
            "additional_instructions",
        ]
    ).set_index(["index_column_1", "index_column_2", "version"])

    actual = utils.construct_generation_data(
        new_keywords_data=new_keywords_data,
        additional_instructions_data=additional_instructions_data,
        n_versions=2,
    )

    expected = pd.DataFrame({
        "index_column_1": ["a", "a", "a", "a", "b", "b"],
        "index_column_2": ["c", "c", "d", "d", "d", "d"],
        "version": ["1", "2", "1", "2", "1", "2"],
        "keywords": [
            "keyword 1",
            "keyword 1",
            "keyword 2",
            "keyword 2",
            "keyword 3, keyword 4",
            "keyword 3, keyword 4",
        ],
        "additional_instructions": ["", "", "", "", "", ""],
        "existing_headlines": [[], [], [], [], [], []],
        "existing_descriptions": [[], [], [], [], [], []],
    }).set_index(["index_column_1", "index_column_2", "version"])

    pd.testing.assert_frame_equal(actual, expected, check_like=True)

  def test_construct_generation_data_with_empty_existing_generations(
      self,
  ):
    new_keywords_data = pd.DataFrame({
        "index_column_1": ["a", "a", "b", "b"],
        "index_column_2": ["c", "d", "d", "d"],
        "keyword": ["keyword 1", "keyword 2", "keyword 3", "keyword 4"],
    }).set_index(["index_column_1", "index_column_2"])

    existing_generations_data = pd.DataFrame(
        columns=[
            "index_column_1",
            "index_column_2",
            "version",
            "existing_headlines",
            "existing_descriptions",
        ]
    ).set_index(["index_column_1", "index_column_2", "version"])

    actual = utils.construct_generation_data(
        new_keywords_data=new_keywords_data,
        existing_generations_data=existing_generations_data,
        n_versions=2,
    )

    expected = pd.DataFrame({
        "index_column_1": ["a", "a", "a", "a", "b", "b"],
        "index_column_2": ["c", "c", "d", "d", "d", "d"],
        "version": ["1", "2", "1", "2", "1", "2"],
        "keywords": [
            "keyword 1",
            "keyword 1",
            "keyword 2",
            "keyword 2",
            "keyword 3, keyword 4",
            "keyword 3, keyword 4",
        ],
        "additional_instructions": ["", "", "", "", "", ""],
        "existing_headlines": [[], [], [], [], [], []],
        "existing_descriptions": [[], [], [], [], [], []],
    }).set_index(["index_column_1", "index_column_2", "version"])

    pd.testing.assert_frame_equal(actual, expected, check_like=True)


class ExplodeGeneratedAdObjectTest(parameterized.TestCase):

  def test_explode_generated_ads_object(self):
    data = pd.DataFrame({
        "generated_ad_object": [
            copycat.CopycatResponse(
                google_ad=google_ads.GoogleAd(
                    headlines=["headline 1", "headline 2"],
                    descriptions=["description 1", "description 2"],
                ),
                keywords="keyword 1, keyword 2",
                evaluation_results=copycat.EvaluationResults(
                    headlines_are_memorised=True,
                    descriptions_are_memorised=True,
                    style_similarity=0.5,
                    keyword_similarity=0.6,
                    warnings=["warning 1", "warning 2"],
                    errors=["error 1", "error 2"],
                ),
            ),
            copycat.CopycatResponse(
                google_ad=google_ads.GoogleAd(
                    headlines=["headline 3", "headline 4"],
                    descriptions=["description 3", "description 4"],
                ),
                keywords="keyword 3, keyword 4",
                evaluation_results=copycat.EvaluationResults(
                    headlines_are_memorised=False,
                    descriptions_are_memorised=False,
                    style_similarity=0.25,
                    keyword_similarity=0.35,
                    warnings=[],
                    errors=[],
                ),
            ),
        ],
    })

    exploded_data = utils.explode_generated_ad_object(data)
    expected_data = pd.DataFrame({
        "headlines": [
            ["headline 1", "headline 2"],
            ["headline 3", "headline 4"],
        ],
        "descriptions": [
            ["description 1", "description 2"],
            ["description 3", "description 4"],
        ],
        "Success": [False, True],
        "Headlines are Memorized": [True, False],
        "Descriptions are Memorized": [True, False],
        "Style Similarity": [0.5, 0.25],
        "Keyword Similarity": [0.6, 0.35],
        "Warnings": ["- warning 1\n- warning 2", ""],
        "Errors": ["- error 1\n- error 2", ""],
        "Headline Count": [2, 2],
        "Description Count": [2, 2],
    })
    expected_data["generated_ad_object"] = data["generated_ad_object"].copy()

    pd.testing.assert_frame_equal(exploded_data, expected_data, check_like=True)


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A python library to generate ad copy that matches an advertisers style."""

from copycat import copycat

Copycat = copycat.Copycat
CopycatResponse = copycat.CopycatResponse
StyleGuideGenerator = copycat.StyleGuideGenerator

GoogleAd = copycat.GoogleAd
GoogleAdFormat = copycat.GoogleAdFormat
ValidationError = copycat.ValidationError
ModelName = copycat.ModelName
EmbeddingModelName = copycat.EmbeddingModelName
TextGenerationRequest = copycat.TextGenerationRequest
ExemplarSelectionMethod = copycat.ExemplarSelectionMethod
EvaluationResults = copycat.EvaluationResults
BirchAgglomerativeKeywordClusterer = copycat.BirchAgglomerativeKeywordClusterer

HarmCategory = copycat.generative_models.HarmCategory
HarmBlockThreshold = copycat.generative_models.HarmBlockThreshold
ALL_SAFETY_SETTINGS_OFF = copycat.ALL_SAFETY_SETTINGS_OFF
ALL_SAFETY_SETTINGS_ONLY_HIGH = copycat.ALL_SAFETY_SETTINGS_ONLY_HIGH


__version__ = "0.0.3"



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mesop as me

from copycat import copycat
from copycat.ui import components
from copycat.ui import event_handlers
from copycat.ui import states
from copycat.ui import styles


_NEW_ADS_PREVIEW_STYLE = me.Style(
    overflow_y="scroll",
    padding=me.Padding.all(20),
    width="100%",
    height=300,
    border=me.Border.all(
        styles.DEFAULT_BORDER_STYLE,
    ),
    border_radius=6,
)


def generation_settings_section(params: states.CopycatParamsState) -> None:
  """Renders the chat model parameters section of the generate new ads page."""
  with components.rounded_box_section(
      title="Generation Settings", width="100%"
  ):
    me.text("Chat Model:")
    me.radio(
        on_change=event_handlers.update_copycat_parameter,
        key="new_ads_chat_model_name",
        options=[
            me.RadioOption(
                label="Gemini 1.5 Flash",
                value=copycat.ModelName.GEMINI_1_5_FLASH.value,
            ),
            me.RadioOption(
                label="Gemini 1.5 Pro",
                value=copycat.ModelName.GEMINI_1_5_PRO.value,
            ),
        ],
        value=params.new_ads_chat_model_name,
        style=me.Style(margin=me.Margin(bottom=10)),
    )
    with me.box(style=me.Style(margin=me.Margin(bottom=20))):
      me.slide_toggle(
          label="Use Style Guide",
          on_change=event_handlers.update_copycat_parameter_from_slide_toggle,
          key="new_ads_use_style_guide",
          checked=params.new_ads_use_style_guide,
      )

    with components.row(width="100%"):
      with me.box(
          style=me.Style(width="45%", margin=me.Margin.symmetric(horizontal=5))
      ):
        me.input(
            label="N In-context Examples",
            on_blur=event_handlers.update_copycat_parameter,
            key="new_ads_num_in_context_examples",
            value=str(params.new_ads_num_in_context_examples),
            appearance="outline",
            type="number",
            style=me.Style(width="100%"),
        )
      with me.box(
          style=me.Style(width="45%", margin=me.Margin.symmetric(horizontal=5))
      ):
        me.input(
            label="Temperature",
            on_blur=event_handlers.update_copycat_parameter,
            key="new_ads_temperature",
            value=str(params.new_ads_temperature),
            appearance="outline",
            type="number",
            style=me.Style(width="100%"),
        )

    with components.row(width="100%"):
      with me.box(
          style=me.Style(width="45%", margin=me.Margin.symmetric(horizontal=5))
      ):
        me.input(
            label="Top-K",
            on_blur=event_handlers.update_copycat_parameter,
            key="new_ads_top_k",
            value=str(params.new_ads_top_k),
            appearance="outline",
            type="number",
            style=me.Style(width="100%"),
        )
      with me.box(
          style=me.Style(width="45%", margin=me.Margin.symmetric(horizontal=5))
      ):
        me.input(
            label="Top-P",
            on_blur=event_handlers.update_copycat_parameter,
            key="new_ads_top_p",
            value=str(params.new_ads_top_p),
            appearance="outline",
            type="number",
            style=me.Style(width="100%"),
        )

    with components.row(gap=10, margin=me.Margin(bottom=20)):
      me.slide_toggle(
          label="Allow memorised headlines",
          on_change=event_handlers.update_copycat_parameter_from_slide_toggle,
          key="new_ads_allow_memorised_headlines",
          checked=params.new_ads_allow_memorised_headlines,
      )
      me.slide_toggle(
          label="Allow memorised descriptions",
          on_change=event_handlers.update_copycat_parameter_from_slide_toggle,
          key="new_ads_allow_memorised_descriptions",
          checked=params.new_ads_allow_memorised_descriptions,
      )

    with components.row():
      with me.box(
          style=me.Style(width="45%", margin=me.Margin.symmetric(horizontal=5))
      ):
        me.input(
            label="Number of Versions per Ad Group",
            on_blur=event_handlers.update_copycat_parameter,
            key="new_ads_number_of_versions",
            value=str(params.new_ads_number_of_versions),
            appearance="outline",
            type="number",
            style=me.Style(width="100%"),
        )


def preview_prompt_section(state: states.AppState) -> None:
  """Renders the preview prompt section of the generate new ads page."""
  with components.rounded_box_section(title="Preview Prompt", width="100%"):
    me.text(
        "Preview the prompt and settings that will be used to generate the"
        " first ad."
    )
    with components.row(
        gap=15,
        align_items="center",
        margin=me.Margin.symmetric(vertical=15),
    ):
      me.button(
          label="Preview",
          type="flat",
          disabled=not state.has_copycat_instance,
          on_click=event_handlers.generate_new_ad_preview,
      )
    if state.new_ad_preview_request:
      with me.box(style=_NEW_ADS_PREVIEW_STYLE):
        me.markdown(state.new_ad_preview_request, style=me.Style(height="40%"))


def generate_new_ads_section(
    params: states.CopycatParamsState, state: states.AppState
) -> None:
  """Renders the generate new ads section of the generate new ads page."""
  with components.rounded_box_section(title="Generate Ads", width="100%"):
    me.text(
        "Copycat will generate ads in batches. The batch size is the number of"
        " ads to generate at a time, and the limit is the total number of ads"
        " to generate. For example, a batch size of 15 and a limit of 25 would"
        " mean two batches are generated, the first 15 ads and then the next 10"
        " ads. The ads will be written to the Google Sheet in a tab named"
        " 'Generated Ads'"
    )
    with components.row(gap=15, align_items="center", margin=me.Margin(top=15)):
      with me.tooltip(
          message=(
              "The batch size for generating ads. A larger batch size will"
              " finish faster but you may hit your quota limit."
          )
      ):
        me.input(
            label="Batch Size",
            value=str(params.new_ads_batch_size),
            type="number",
            appearance="outline",
            style=me.Style(width=100),
            on_blur=event_handlers.update_copycat_parameter,
            key="new_ads_batch_size",
        )
      with me.tooltip(
          message=(
              "The maximum number of ads to generate. If 0 then there is no"
              " limit."
          )
      ):
        me.input(
            label="Limit",
            value=str(params.new_ads_generation_limit),
            type="number",
            appearance="outline",
            style=me.Style(width=100),
            on_blur=event_handlers.update_copycat_parameter,
            key="new_ads_generation_limit",
        )
    with components.row(gap=15, align_items="center", margin=me.Margin(top=15)):
      me.button(
          label="Generate",
          type="flat",
          disabled=not state.has_copycat_instance,
          on_click=event_handlers.generate_ads,
      )
      me.slide_toggle(
          label="Fill Gaps",
          on_change=event_handlers.update_copycat_parameter_from_slide_toggle,
          key="new_ads_fill_gaps",
          checked=params.new_ads_fill_gaps,
      )


def generate_new_ads() -> None:
  """Renders the generate new ads page."""
  state = me.state(states.AppState)
  params = me.state(states.CopycatParamsState)

  with components.column(gap=15, width="50%"):
    with components.row(width="100%"):
      generation_settings_section(params)

  with components.column(gap=15, width="50%"):
    with components.row(width="100%"):
      preview_prompt_section(state)

    with components.row(width="100%"):
      generate_new_ads_section(params, state)



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import keyword

from absl.testing import absltest
from absl.testing import parameterized
from vertexai import generative_models
import pandas as pd

from copycat import copycat
from copycat import google_ads
from copycat import testing_utils


class CopycatResponseTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    self.training_headlines = ["train headline 1", "train headline 2"]
    self.training_descriptions = ["train description 1", "train description 2"]
    self.keywords = "keyword 1, keyword 2"

    self.default_kwargs = dict(
        training_headlines=self.training_headlines,
        training_descriptions=self.training_descriptions,
        keywords=self.keywords,
    )

  @parameterized.parameters([([], True), (["Non empty error message"], False)])
  def test_success_is_false_if_there_is_an_error_message(
      self, errors, expected_success
  ):
    generated_google_ad = google_ads.GoogleAd(
        headlines=["New headline 1", "New headline 2"],
        descriptions=["New description"],
    )

    response = copycat.CopycatResponse(
        google_ad=generated_google_ad,
        keywords=self.keywords,
        evaluation_results=copycat.EvaluationResults(
            headlines_are_memorised=False,
            descriptions_are_memorised=False,
            style_similarity=None,
            keyword_similarity=None,
            errors=errors,
            warnings=[],
        ),
    )
    self.assertEqual(response.success, expected_success)

  @parameterized.parameters([
      ([], ""),
      (["One error message"], "- One error message"),
      (
          ["Two error message", "Two error message (b)"],
          "- Two error message\n- Two error message (b)",
      ),
  ])
  def test_error_message_is_created_by_joining_errors(
      self, errors, expected_error_message
  ):
    generated_google_ad = google_ads.GoogleAd(
        headlines=["New headline 1", "New headline 2"],
        descriptions=["New description"],
    )

    response = copycat.CopycatResponse(
        google_ad=generated_google_ad,
        keywords=self.keywords,
        evaluation_results=copycat.EvaluationResults(
            headlines_are_memorised=False,
            descriptions_are_memorised=False,
            style_similarity=None,
            keyword_similarity=None,
            errors=errors,
            warnings=[],
        ),
    )
    self.assertEqual(response.error_message, expected_error_message)

  @parameterized.parameters([
      ([], ""),
      (["One warning message"], "- One warning message"),
      (
          ["Two warning message", "Two warning message (b)"],
          "- Two warning message\n- Two warning message (b)",
      ),
  ])
  def test_warning_message_is_created_by_joining_warnings(
      self, warnings, expected_warning_message
  ):
    generated_google_ad = google_ads.GoogleAd(
        headlines=["New headline 1", "New headline 2"],
        descriptions=["New description"],
    )

    response = copycat.CopycatResponse(
        google_ad=generated_google_ad,
        keywords=self.keywords,
        evaluation_results=copycat.EvaluationResults(
            headlines_are_memorised=False,
            descriptions_are_memorised=False,
            style_similarity=None,
            keyword_similarity=None,
            warnings=warnings,
            errors=[],
        ),
    )
    self.assertEqual(response.warning_message, expected_warning_message)

  def test_raise_if_not_success_raises_exception_if_not_success(self):
    response = copycat.CopycatResponse(
        google_ad=google_ads.GoogleAd(
            headlines=["New headline 1", "New headline 2"],
            descriptions=["New description"],
        ),
        keywords=self.keywords,
        evaluation_results=copycat.EvaluationResults(
            headlines_are_memorised=False,
            descriptions_are_memorised=False,
            style_similarity=None,
            keyword_similarity=None,
            errors=["One error message"],
            warnings=[],
        ),
    )
    with self.assertRaisesWithLiteralMatch(
        copycat.CopycatResponseError, "- One error message"
    ):
      response.raise_if_not_success()

  def test_raise_if_not_success_does_not_raise_if_success(self):
    response = copycat.CopycatResponse(
        google_ad=google_ads.GoogleAd(
            headlines=["New headline 1", "New headline 2"],
            descriptions=["New description"],
        ),
        keywords=self.keywords,
        evaluation_results=copycat.EvaluationResults(
            headlines_are_memorised=False,
            descriptions_are_memorised=False,
            style_similarity=None,
            keyword_similarity=None,
            errors=[],
            warnings=[],
        ),
    )
    response.raise_if_not_success()


class CopycatTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    self.training_data = lambda n_rows: pd.DataFrame({
        "headlines": [
            [f"train headline {i}", f"train headline {i} (2)"]
            for i in range(1, n_rows + 1)
        ],
        "descriptions": [
            [f"train description {i}"] for i in range(1, n_rows + 1)
        ],
        "keywords": [
            f"keyword {i}a, keyword {i}b" for i in range(1, n_rows + 1)
        ],
        "redundant_column": [
            f"redundant col {i}" for i in range(1, n_rows + 1)
        ],
    })
    self.tmp_dir = self.create_tempdir()

    self.model_name = "gemini-1.5-flash-002"

    self.embedding_model_patcher = testing_utils.PatchEmbeddingsModel()
    self.embedding_model_patcher.start()

  def tearDown(self):
    super().tearDown()
    self.embedding_model_patcher.stop()

  def test_create_from_pandas_creates_copycat_instance(
      self,
  ):
    copycat_instance = copycat.Copycat.create_from_pandas(
        training_data=self.training_data(n_rows=3),
        embedding_model_name="text-embedding-004",
        ad_format="text_ad",
        vectorstore_exemplar_selection_method="random",
    )

    self.assertIsInstance(copycat_instance, copycat.Copycat)

  def test_create_from_pandas_uses_training_data_to_construct_vectorstore(
      self,
  ):
    copycat_instance = copycat.Copycat.create_from_pandas(
        training_data=self.training_data(n_rows=3),
        embedding_model_name="text-embedding-004",
        ad_format="text_ad",
        vectorstore_exemplar_selection_method="random",
    )

    # Will return all the entries because we only had 3 rows in the training
    # data.
    vectorstore_results = copycat_instance.ad_copy_vectorstore.get_relevant_ads(
        ["query"], k=3
    )[0]

    expected_vectorstore_results = [
        copycat.ad_copy_generator.ExampleAd(
            keywords="keyword 1a, keyword 1b",
            google_ad=google_ads.GoogleAd(
                headlines=["train headline 1", "train headline 1 (2)"],
                descriptions=["train description 1"],
            ),
        ),
        copycat.ad_copy_generator.ExampleAd(
            keywords="keyword 2a, keyword 2b",
            google_ad=google_ads.GoogleAd(
                headlines=["train headline 2", "train headline 2 (2)"],
                descriptions=["train description 2"],
            ),
        ),
        copycat.ad_copy_generator.ExampleAd(
            keywords="keyword 3a, keyword 3b",
            google_ad=google_ads.GoogleAd(
                headlines=["train headline 3", "train headline 3 (2)"],
                descriptions=["train description 3"],
            ),
        ),
    ]
    self.assertCountEqual(vectorstore_results, expected_vectorstore_results)

  def test_create_from_pandas_replaces_special_variables_in_training_data_if_set(
      self,
  ):
    training_data = pd.DataFrame({
        "headlines": [
            ["headline 1", "headline {CUSTOMIZER.product:my product}"]
        ],
        "descriptions": [
            ["description 1", "description {CUSTOMIZER.product:my product}"]
        ],
        "keywords": ["keyword 1, keyword 2"],
    })
    copycat_instance = copycat.Copycat.create_from_pandas(
        training_data=training_data,
        embedding_model_name="text-embedding-004",
        ad_format="text_ad",
        vectorstore_exemplar_selection_method="random",
        replace_special_variables_with_default=True,
    )

    # Will return all the entries because we only had 3 rows in the training
    # data.
    vectorstore_results = copycat_instance.ad_copy_vectorstore.get_relevant_ads(
        ["query"], k=3
    )[0]

    expected_vectorstore_results = [
        copycat.ad_copy_generator.ExampleAd(
            keywords="keyword 1, keyword 2",
            google_ad=google_ads.GoogleAd(
                headlines=["headline 1", "headline my product"],
                descriptions=["description 1", "description my product"],
            ),
        ),
    ]
    self.assertCountEqual(vectorstore_results, expected_vectorstore_results)

  def test_create_from_pandas_considers_unfillable_special_variables_as_invalid_ads_if_replacing_them(
      self,
  ):
    training_data = pd.DataFrame({
        "headlines": [["headline 1", "headline {CUSTOMIZER.product}"]],
        "descriptions": [["description 1", "description {CUSTOMIZER.product}"]],
        "keywords": ["keyword 1, keyword 2"],
    })
    with self.assertRaisesWithLiteralMatch(
        ValueError, "1 (100.00%) invalid ads found in the training data."
    ):
      copycat.Copycat.create_from_pandas(
          training_data=training_data,
          embedding_model_name="text-embedding-004",
          ad_format="text_ad",
          vectorstore_exemplar_selection_method="random",
          replace_special_variables_with_default=True,
          on_invalid_ad="raise",
      )

  def test_create_from_pandas_ignores_special_variables_in_training_data_if_not_replacing_them(
      self,
  ):
    training_data = pd.DataFrame({
        "headlines": [
            ["headline 1", "headline {CUSTOMIZER.product:my product}"]
        ],
        "descriptions": [["description 1", "description {CUSTOMIZER.product}"]],
        "keywords": ["keyword 1, keyword 2"],
    })
    copycat_instance = copycat.Copycat.create_from_pandas(
        training_data=training_data,
        embedding_model_name="text-embedding-004",
        ad_format="text_ad",
        vectorstore_exemplar_selection_method="random",
        replace_special_variables_with_default=False,
    )

    # Will return all the entries because we only had 3 rows in the training
    # data.
    vectorstore_results = copycat_instance.ad_copy_vectorstore.get_relevant_ads(
        ["query"], k=3
    )[0]

    expected_vectorstore_results = [
        copycat.ad_copy_generator.ExampleAd(
            keywords="keyword 1, keyword 2",
            google_ad=google_ads.GoogleAd(
                headlines=[
                    "headline 1",
                    "headline {CUSTOMIZER.product:my product}",
                ],
                descriptions=[
                    "description 1",
                    "description {CUSTOMIZER.product}",
                ],
            ),
        ),
    ]
    self.assertCountEqual(vectorstore_results, expected_vectorstore_results)

  def test_create_from_pandas_on_invalid_ad_is_skip_ignores_invalid_ads(self):
    training_data = pd.DataFrame.from_records([
        {
            "headlines": ["a" * 31, "invalid headline"],
            "descriptions": ["invalid description 1"],
            "keywords": "keyword 1, keyword 2",
        },
        {
            "headlines": ["headline 3"],
            "descriptions": ["description 3"],
            "keywords": "keyword 3, keyword 4",
        },
    ])

    with self.assertWarnsRegex(
        Warning,
        "^1 \(50\.00%\) invalid ads found in the training data\. Keeping them"
        " in the training data\.$",
    ):
      copycat_instance = copycat.Copycat.create_from_pandas(
          training_data=training_data,
          embedding_model_name="text-embedding-004",
          ad_format="text_ad",
          on_invalid_ad="skip",
          vectorstore_exemplar_selection_method="random",
      )

    pd.testing.assert_frame_equal(
        copycat_instance.ad_copy_vectorstore.ad_exemplars[
            ["headlines", "descriptions", "keywords"]
        ],
        training_data,
        check_like=True,
    )

  def test_create_from_pandas_on_invalid_ad_is_raise_raises_exception(self):
    training_data = pd.DataFrame.from_records([
        {
            "headlines": ["a" * 31, "invalid headline"],
            "descriptions": ["invalid description 1"],
            "keywords": "keyword 1, keyword 2",
        },
        {
            "headlines": ["headline 3"],
            "descriptions": ["description 3"],
            "keywords": "keyword 3, keyword 4",
        },
    ])

    with self.assertRaisesRegex(
        ValueError,
        "^1 \(50\.00%\) invalid ads found in the training data\.",
    ):
      copycat.Copycat.create_from_pandas(
          training_data=training_data,
          embedding_model_name="text-embedding-004",
          ad_format="text_ad",
          on_invalid_ad="raise",
          vectorstore_exemplar_selection_method="random",
      )

  def test_create_from_pandas_on_invalid_ad_is_drop_drops_invalid_ads(self):
    training_data = pd.DataFrame.from_records([
        {
            "headlines": ["a" * 31, "invalid headline"],
            "descriptions": ["invalid description 1"],
            "keywords": "keyword 1, keyword 2",
        },
        {
            "headlines": ["headline 3"],
            "descriptions": ["description 3"],
            "keywords": "keyword 3, keyword 4",
        },
    ])

    with self.assertWarnsRegex(
        Warning,
        "^1 \(50\.00%\) invalid ads found in the training data\. Dropping them"
        " from the training data\.$",
    ):
      copycat_instance = copycat.Copycat.create_from_pandas(
          training_data=training_data,
          embedding_model_name="text-embedding-004",
          ad_format="text_ad",
          on_invalid_ad="drop",
          vectorstore_exemplar_selection_method="random",
      )

    pd.testing.assert_frame_equal(
        copycat_instance.ad_copy_vectorstore.ad_exemplars[
            ["headlines", "descriptions", "keywords"]
        ],
        pd.DataFrame.from_records([
            {
                "headlines": ["headline 3"],
                "descriptions": ["description 3"],
                "keywords": "keyword 3, keyword 4",
            },
        ]),
        check_like=True,
    )

  @parameterized.parameters("headlines", "descriptions", "keywords")
  def test_create_from_pandas_raises_exception_when_required_column_is_missing(
      self, missing_column
  ):
    expected_error_message = (
        "Training data must contain the columns ['descriptions',"
        f" 'headlines', 'keywords']. Missing columns: ['{missing_column}']."
    )
    with self.assertRaisesWithLiteralMatch(ValueError, expected_error_message):
      copycat.Copycat.create_from_pandas(
          training_data=self.training_data(3).drop(columns=[missing_column]),
          embedding_model_name="text-embedding-004",
          ad_format="text_ad",
          vectorstore_exemplar_selection_method="random",
      )

  @parameterized.named_parameters(
      dict(
          testcase_name="without keyword specific instructions",
          keywords="my keyword 1, my keyword 2",
          existing_headlines=None,
          existing_descriptions=None,
          keywords_specific_instructions="",
          expected_prompt=[
              generative_models.Content(
                  role="user",
                  parts=[
                      generative_models.Part.from_text(
                          "Please write 2 headlines and 1 descriptions for this"
                          " ad.\n\nKeywords: keyword 11a, keyword 11b"
                      )
                  ],
              ),
              generative_models.Content(
                  role="model",
                  parts=[
                      generative_models.Part.from_text(
                          '{"headlines":["train headline 11","train headline 11'
                          ' (2)"],"descriptions":["train description 11"]}'
                      )
                  ],
              ),
              generative_models.Content(
                  role="user",
                  parts=[
                      generative_models.Part.from_text(
                          "Please write 2 headlines and 1 descriptions for this"
                          " ad.\n\nKeywords: keyword 20a, keyword 20b"
                      )
                  ],
              ),
              generative_models.Content(
                  role="model",
                  parts=[
                      generative_models.Part.from_text(
                          '{"headlines":["train headline 20","train headline 20'
                          ' (2)"],"descriptions":["train description 20"]}'
                      )
                  ],
              ),
              generative_models.Content(
                  role="user",
                  parts=[
                      generative_models.Part.from_text(
                          "Please write 3 headlines and 2 descriptions for this"
                          " ad.\n\nKeywords: my keyword 1, my keyword 2"
                      )
                  ],
              ),
          ],
      ),
      dict(
          testcase_name="with keyword specific instructions",
          keywords="my keyword 1, my keyword 2",
          existing_headlines=None,
          existing_descriptions=None,
          keywords_specific_instructions="Some keyword specific instructions.",
          expected_prompt=[
              generative_models.Content(
                  role="user",
                  parts=[
                      generative_models.Part.from_text(
                          "Please write 2 headlines and 1 descriptions for this"
                          " ad.\n\nKeywords: keyword 11a, keyword 11b"
                      )
                  ],
              ),
              generative_models.Content(
                  role="model",
                  parts=[
                      generative_models.Part.from_text(
                          '{"headlines":["train headline 11","train headline 11'
                          ' (2)"],"descriptions":["train description 11"]}'
                      )
                  ],
              ),
              generative_models.Content(
                  role="user",
                  parts=[
                      generative_models.Part.from_text(
                          "Please write 2 headlines and 1 descriptions for this"
                          " ad.\n\nKeywords: keyword 20a, keyword 20b"
                      )
                  ],
              ),
              generative_models.Content(
                  role="model",
                  parts=[
                      generative_models.Part.from_text(
                          '{"headlines":["train headline 20","train headline 20'
                          ' (2)"],"descriptions":["train description 20"]}'
                      )
                  ],
              ),
              generative_models.Content(
                  role="user",
                  parts=[
                      generative_models.Part.from_text(
                          f"For the next set of keywords, please consider the"
                          f" following additional instructions:\n\nSome keyword"
                          f" specific instructions.\n\nPlease write 3 headlines"
                          f" and 2 descriptions for this ad.\n\nKeywords: my"
                          f" keyword 1, my keyword 2"
                      )
                  ],
              ),
          ],
      ),
      dict(
          testcase_name="with existing ad copy",
          keywords="my keyword 1, my keyword 2",
          existing_headlines=["existing headline"],
          existing_descriptions=["existing description"],
          keywords_specific_instructions="",
          expected_prompt=[
              generative_models.Content(
                  role="user",
                  parts=[
                      generative_models.Part.from_text(
                          "Please write 2 headlines and 1 descriptions for this"
                          " ad.\n\nKeywords: keyword 11a, keyword 11b"
                      )
                  ],
              ),
              generative_models.Content(
                  role="model",
                  parts=[
                      generative_models.Part.from_text(
                          '{"headlines":["train headline 11","train headline 11'
                          ' (2)"],"descriptions":["train description 11"]}'
                      )
                  ],
              ),
              generative_models.Content(
                  role="user",
                  parts=[
                      generative_models.Part.from_text(
                          "Please write 2 headlines and 1 descriptions for this"
                          " ad.\n\nKeywords: keyword 20a, keyword 20b"
                      )
                  ],
              ),
              generative_models.Content(
                  role="model",
                  parts=[
                      generative_models.Part.from_text(
                          '{"headlines":["train headline 20","train headline 20'
                          ' (2)"],"descriptions":["train description 20"]}'
                      )
                  ],
              ),
              generative_models.Content(
                  role="user",
                  parts=[
                      generative_models.Part.from_text(
                          "This ad already has 1 headlines and 1"
                          " descriptions:\n\n- headlines: ['existing"
                          " headline']\n- descriptions: ['existing"
                          " description']\n\nPlease write 2 more headlines and"
                          " 1 more descriptions to complete this"
                          " ad.\n\nKeywords: my keyword 1, my keyword 2"
                      )
                  ],
              ),
          ],
      ),
  )
  def test_construct_text_generation_requests_for_new_ad_copy_returns_expected_request(
      self,
      keywords,
      existing_headlines,
      existing_descriptions,
      keywords_specific_instructions,
      expected_prompt,
  ):
    copycat_instance = copycat.Copycat.create_from_pandas(
        training_data=self.training_data(n_rows=20),
        embedding_model_name="text-embedding-004",
        ad_format="text_ad",
        vectorstore_exemplar_selection_method="random",
    )
    request = (
        copycat_instance.construct_text_generation_requests_for_new_ad_copy(
            keywords=[keywords],
            existing_headlines=[existing_headlines],
            existing_descriptions=[existing_descriptions],
            keywords_specific_instructions=[keywords_specific_instructions],
            num_in_context_examples=2,
            system_instruction="Example system instruction",
        )[0]
    )

    expected_request = copycat.TextGenerationRequest(
        keywords=keywords,
        prompt=expected_prompt,
        system_instruction="Example system instruction",
        chat_model_name=copycat.ModelName.GEMINI_1_5_FLASH,
        temperature=0.95,
        top_k=20,
        top_p=0.95,
        safety_settings=None,
        existing_ad_copy=google_ads.GoogleAd(
            headlines=existing_headlines or [],
            descriptions=existing_descriptions or [],
        ),
    )
    self.assertEqual(expected_request.to_markdown(), request.to_markdown())

  @parameterized.named_parameters(
      dict(
          testcase_name="with existing ad copy",
          existing_headlines=["existing headline"],
          existing_descriptions=["existing description"],
          expected_style_similarity=0.5590658228752543,
          expected_keyword_similarity=0.5056089825975687,
      ),
      dict(
          testcase_name="without existing ad copy",
          existing_headlines=None,
          existing_descriptions=None,
          expected_style_similarity=0.5338446522650305,
          expected_keyword_similarity=0.4766861086664519,
      ),
  )
  @testing_utils.PatchGenerativeModel(
      response=(
          '{"headlines": ["generated headline 1", "generated headline 2"],'
          ' "descriptions": ["generated description"]}'
      )
  )
  def test_generate_new_ad_copy_returns_expected_response(
      self,
      generative_model_patcher,
      existing_headlines,
      existing_descriptions,
      expected_style_similarity,
      expected_keyword_similarity,
  ):

    copycat_instance = copycat.Copycat.create_from_pandas(
        training_data=self.training_data(20),
        embedding_model_name="text-embedding-004",
        ad_format="text_ad",
        vectorstore_exemplar_selection_method="random",
    )
    response = copycat_instance.generate_new_ad_copy(
        keywords=["my keyword 1, my keyword 2"],
        style_guide="This is my style guide.",
        num_in_context_examples=2,
        system_instruction_kwargs=dict(
            company_name="My company",
            language="english",
        ),
        existing_headlines=[existing_headlines],
        existing_descriptions=[existing_descriptions],
    )[0]

    existing_headlines = existing_headlines or []
    existing_descriptions = existing_descriptions or []
    generated_headlines = ["generated headline 1", "generated headline 2"]
    generated_descriptions = ["generated description"]

    self.assertEqual(
        response,
        copycat.CopycatResponse(
            google_ad=google_ads.GoogleAd(
                headlines=existing_headlines + generated_headlines,
                descriptions=existing_descriptions + generated_descriptions,
            ),
            keywords="my keyword 1, my keyword 2",
            evaluation_results=copycat.EvaluationResults(
                errors=[],
                warnings=[],
                headlines_are_memorised=False,
                descriptions_are_memorised=False,
                style_similarity=expected_style_similarity,
                keyword_similarity=expected_keyword_similarity,
            ),
        ),
    )

  @testing_utils.PatchGenerativeModel(response="Generated style guide.")
  def test_generate_style_guide_returns_style_guide_and_updates_attribute(
      self, generative_model_patcher
  ):

    copycat_instance = copycat.Copycat.create_from_pandas(
        training_data=self.training_data(20),
        embedding_model_name="text-embedding-004",
        ad_format="text_ad",
        vectorstore_exemplar_selection_method="random",
    )
    style_guide = copycat_instance.generate_style_guide(
        company_name="My company"
    )
    self.assertEqual(style_guide, "Generated style guide.")
    self.assertEqual(copycat_instance.style_guide, "Generated style guide.")

  def test_generate_style_guide_raises_exception_if_model_fails_to_generate(
      self,
  ):
    failed_response = generative_models.GenerationResponse.from_dict({
        "candidates": [{
            "finish_reason": generative_models.FinishReason.SAFETY,
        }]
    })

    with testing_utils.PatchGenerativeModel(response=failed_response):
      copycat_instance = copycat.Copycat.create_from_pandas(
          training_data=self.training_data(20),
          embedding_model_name="text-embedding-004",
          ad_format="text_ad",
          vectorstore_exemplar_selection_method="random",
      )

      with self.assertRaises(RuntimeError):
        copycat_instance.generate_style_guide(
            company_name="My company",
            additional_style_instructions="Some additional style instructions.",
            model_name="gemini-1.5-flash-001",
            temperature=0.95,
            top_k=20,
            top_p=0.95,
        )

  @parameterized.named_parameters(
      dict(testcase_name="empty company name", args=dict(company_name="")),
      dict(
          testcase_name="no exemplars or files uri provided",
          args=dict(
              company_name="My company", use_exemplar_ads=False, files_uri=""
          ),
      ),
  )
  @testing_utils.PatchGenerativeModel(response="Generated style guide.")
  def test_generate_style_guide_raises_exception_for_bad_args(
      self, generative_model_patcher, args
  ):

    copycat_instance = copycat.Copycat.create_from_pandas(
        training_data=self.training_data(20),
        embedding_model_name="text-embedding-004",
        ad_format="text_ad",
        vectorstore_exemplar_selection_method="random",
    )

    with self.assertRaises(ValueError):
      copycat_instance.generate_style_guide(**args)

  @parameterized.named_parameters(
      dict(
          testcase_name="no style guide",
          generate_style_guide=False,
          style_guide_arg=None,
          expected_system_instruction="Example system instruction",
      ),
      dict(
          testcase_name="pass style guide",
          generate_style_guide=False,
          style_guide_arg="This is my style guide.",
          expected_system_instruction=(
              "Example system instruction\n\nThis is my style guide."
          ),
      ),
      dict(
          testcase_name="generated style guide",
          generate_style_guide=True,
          style_guide_arg=None,
          expected_system_instruction=(
              "Example system instruction\n\nGenerated style guide."
          ),
      ),
      dict(
          testcase_name="generated style guide overridden",
          generate_style_guide=True,
          style_guide_arg="This is my style guide.",
          expected_system_instruction=(
              "Example system instruction\n\nThis is my style guide."
          ),
      ),
      dict(
          testcase_name="generated style guide overridden with nothing",
          generate_style_guide=True,
          style_guide_arg="",
          expected_system_instruction="Example system instruction",
      ),
  )
  @testing_utils.PatchGenerativeModel(response="Generated style guide.")
  def test_construct_text_generation_requests_uses_generated_style_guide(
      self,
      generative_model_patcher,
      generate_style_guide,
      style_guide_arg,
      expected_system_instruction,
  ):
    copycat_instance = copycat.Copycat.create_from_pandas(
        training_data=self.training_data(n_rows=20),
        embedding_model_name="text-embedding-004",
        ad_format="text_ad",
        vectorstore_exemplar_selection_method="random",
    )
    if generate_style_guide:
      copycat_instance.generate_style_guide(
          company_name="My company",
          additional_style_instructions="Some additional style instructions.",
          model_name="gemini-1.5-flash-001",
          temperature=0.95,
          top_k=20,
          top_p=0.95,
      )

    request = (
        copycat_instance.construct_text_generation_requests_for_new_ad_copy(
            keywords=["my keyword 1, my keyword 2"],
            existing_headlines=[["Headline 1"]],
            existing_descriptions=[["Description 1"]],
            keywords_specific_instructions=[""],
            num_in_context_examples=1,
            system_instruction="Example system instruction",
            style_guide=style_guide_arg,
        )[0]
    )
    self.assertEqual(expected_system_instruction, request.system_instruction)

  @testing_utils.PatchGenerativeModel(
      response='{"descriptions": ["generated description"]}'
  )
  def test_generate_new_ad_copy_returns_expected_response_if_headlines_is_missing(
      self,
      generative_model_patcher,
  ):

    copycat_instance = copycat.Copycat.create_from_pandas(
        training_data=self.training_data(20),
        embedding_model_name="text-embedding-004",
        ad_format="text_ad",
        vectorstore_exemplar_selection_method="random",
    )
    response = copycat_instance.generate_new_ad_copy(
        keywords=["my keyword 1, my keyword 2"],
        style_guide="This is my style guide.",
        num_in_context_examples=2,
        system_instruction_kwargs=dict(
            company_name="My company",
            language="english",
        ),
    )[0]

    # I don't want to test the similarity metrics here, so I'm just setting
    # them to None.
    response.evaluation_results.style_similarity = None
    response.evaluation_results.keyword_similarity = None

    self.assertEqual(
        response,
        copycat.CopycatResponse(
            google_ad=google_ads.GoogleAd(
                headlines=[],
                descriptions=["generated description"],
            ),
            keywords="my keyword 1, my keyword 2",
            evaluation_results=copycat.EvaluationResults(
                errors=["Invalid number of headlines for the ad format."],
                warnings=[],
                headlines_are_memorised=False,
                descriptions_are_memorised=False,
                style_similarity=None,
                keyword_similarity=None,
            ),
        ),
    )

  @testing_utils.PatchGenerativeModel(
      response='{"headlines": ["generated headline"]}'
  )
  def test_generate_new_ad_copy_returns_expected_response_if_descriptions_is_missing(
      self,
      generative_model_patcher,
  ):

    copycat_instance = copycat.Copycat.create_from_pandas(
        training_data=self.training_data(20),
        embedding_model_name="text-embedding-004",
        ad_format="text_ad",
        vectorstore_exemplar_selection_method="random",
    )
    response = copycat_instance.generate_new_ad_copy(
        keywords=["my keyword 1, my keyword 2"],
        style_guide="This is my style guide.",
        num_in_context_examples=2,
        system_instruction_kwargs=dict(
            company_name="My company",
            language="english",
        ),
    )[0]

    # I don't want to test the similarity metrics here, so I'm just setting
    # them to None.
    response.evaluation_results.style_similarity = None
    response.evaluation_results.keyword_similarity = None

    self.assertEqual(
        response,
        copycat.CopycatResponse(
            google_ad=google_ads.GoogleAd(
                headlines=["generated headline"],
                descriptions=[],
            ),
            keywords="my keyword 1, my keyword 2",
            evaluation_results=copycat.EvaluationResults(
                errors=["Invalid number of descriptions for the ad format."],
                warnings=[],
                headlines_are_memorised=False,
                descriptions_are_memorised=False,
                style_similarity=None,
                keyword_similarity=None,
            ),
        ),
    )

  @testing_utils.PatchGenerativeModel(
      response=(
          '{"headlines": ["generated headline 1", "generated headline 2"],'
          ' "descriptions": ["generated description"]}'
      )
  )
  def test_generate_new_ad_copy_returns_expected_responses_for_list_of_keywords(
      self, generative_model_patcher
  ):

    copycat_instance = copycat.Copycat.create_from_pandas(
        training_data=self.training_data(20),
        embedding_model_name="text-embedding-004",
        ad_format="text_ad",
        vectorstore_exemplar_selection_method="random",
    )
    responses = copycat_instance.generate_new_ad_copy(
        keywords=["my keyword 1, my keyword 2", "another keyword"],
        style_guide="This is my style guide.",
        num_in_context_examples=2,
        system_instruction_kwargs=dict(
            company_name="My company",
            language="english",
        ),
    )
    self.assertLen(responses, 2)

  @parameterized.named_parameters(
      dict(
          testcase_name="with existing ad copy",
          existing_headlines=["existing headline"],
          existing_descriptions=["existing description"],
      ),
      dict(
          testcase_name="without existing ad copy",
          existing_headlines=None,
          existing_descriptions=None,
      ),
  )
  @testing_utils.PatchGenerativeModel(response="not a json")
  def test_generate_new_ad_copy_returns_expected_response_for_non_json_chat_model_output(
      self, generative_model_patcher, existing_headlines, existing_descriptions
  ):

    copycat_instance = copycat.Copycat.create_from_pandas(
        training_data=self.training_data(20),
        embedding_model_name="text-embedding-004",
        ad_format="text_ad",
        vectorstore_exemplar_selection_method="random",
    )

    response = copycat_instance.generate_new_ad_copy(
        keywords=["my keyword 1, my keyword 2"],
        style_guide="This is my style guide.",
        num_in_context_examples=2,
        system_instruction_kwargs=dict(
            company_name="My company",
            language="english",
        ),
        existing_headlines=[existing_headlines],
        existing_descriptions=[existing_descriptions],
    )[0]

    # I don't want to test the similarity metrics here, so I'm just setting
    # them to None.
    response.evaluation_results.style_similarity = None
    response.evaluation_results.keyword_similarity = None

    self.assertEqual(
        response,
        copycat.CopycatResponse(
            keywords="my keyword 1, my keyword 2",
            google_ad=google_ads.GoogleAd(
                headlines=existing_headlines or [],
                descriptions=existing_descriptions or [],
            ),
            evaluation_results=copycat.EvaluationResults(
                errors=[
                    "1 validation error for GoogleAd\n  Invalid JSON: expected"
                    " ident at line 1 column 2 [type=json_invalid,"
                    " input_value='not a json', input_type=str]\n    For"
                    " further information visit"
                    " https://errors.pydantic.dev/2.6/v/json_invalid"
                ],
                warnings=[],
                headlines_are_memorised=False if existing_headlines else None,
                descriptions_are_memorised=False
                if existing_descriptions
                else None,
                style_similarity=None,
                keyword_similarity=None,
            ),
        ),
    )

  @parameterized.named_parameters(
      dict(
          testcase_name="all descriptions too long",
          headlines=["generated headline"],
          descriptions=["a" * 91],
          expected_error_message=(
              "Invalid number of descriptions for the ad format."
          ),
          expected_headlines_are_memorised=False,
          expected_descriptions_are_memorised=False,
          expected_headlines=["generated headline"],
          expected_descriptions=[],
      ),
      dict(
          testcase_name="all headlines too long",
          headlines=["a" * 31],
          descriptions=["generated description"],
          expected_error_message=(
              "Invalid number of headlines for the ad format."
          ),
          expected_headlines_are_memorised=False,
          expected_descriptions_are_memorised=False,
          expected_headlines=[],
          expected_descriptions=["generated description"],
      ),
      dict(
          testcase_name="headline is memorised",
          headlines=["train headline 1"],
          descriptions=["generated description"],
          expected_error_message=(
              "All headlines are memorised from the training data."
          ),
          expected_headlines_are_memorised=True,
          expected_descriptions_are_memorised=False,
          expected_headlines=["train headline 1"],
          expected_descriptions=["generated description"],
      ),
      dict(
          testcase_name="description is memorised",
          headlines=["generated headline"],
          descriptions=["train description 1"],
          expected_error_message=(
              "All descriptions are memorised from the training data."
          ),
          expected_headlines_are_memorised=False,
          expected_descriptions_are_memorised=True,
          expected_headlines=["generated headline"],
          expected_descriptions=["train description 1"],
      ),
  )
  def test_generate_new_ad_copy_returns_expected_response_if_chat_model_generates_bad_headlines_or_descriptions(
      self,
      headlines,
      descriptions,
      expected_error_message,
      expected_headlines_are_memorised,
      expected_descriptions_are_memorised,
      expected_headlines,
      expected_descriptions,
  ):

    with testing_utils.PatchGenerativeModel(
        response=json.dumps({
            "headlines": headlines,
            "descriptions": descriptions,
        })
    ):
      copycat_instance = copycat.Copycat.create_from_pandas(
          training_data=self.training_data(20),
          embedding_model_name="text-embedding-004",
          ad_format="text_ad",
          vectorstore_exemplar_selection_method="random",
      )

      response = copycat_instance.generate_new_ad_copy(
          keywords=["my keyword 1, my keyword 2"],
          style_guide="This is my style guide.",
          num_in_context_examples=2,
          system_instruction_kwargs=dict(
              company_name="My company",
              language="english",
          ),
          allow_memorised_headlines=False,
          allow_memorised_descriptions=False,
      )[0]

      # I don't want to test the similarity metrics here, so I'm just setting
      # them to None.
      response.evaluation_results.style_similarity = None
      response.evaluation_results.keyword_similarity = None

      expected_google_ad = google_ads.GoogleAd(
          headlines=expected_headlines,
          descriptions=expected_descriptions,
      )
      expected_response = copycat.CopycatResponse(
          google_ad=expected_google_ad,
          keywords="my keyword 1, my keyword 2",
          evaluation_results=copycat.EvaluationResults(
              errors=[expected_error_message],
              warnings=[],
              headlines_are_memorised=expected_headlines_are_memorised,
              descriptions_are_memorised=expected_descriptions_are_memorised,
              style_similarity=None,
              keyword_similarity=None,
          ),
      )

      self.assertEqual(response, expected_response)

  @parameterized.named_parameters(
      dict(
          testcase_name="with existing ad copy",
          existing_headlines=["existing headline"],
          existing_descriptions=["existing description"],
          expected_style_similarity=0.569982855557986,
          expected_keyword_similarity=0.49631553094971914,
      ),
      dict(
          testcase_name="without existing ad copy",
          existing_headlines=None,
          existing_descriptions=None,
          expected_style_similarity=None,
          expected_keyword_similarity=None,
      ),
  )
  def test_generate_new_ad_copy_returns_expected_response_if_chat_model_fails_to_generate(
      self,
      existing_headlines,
      existing_descriptions,
      expected_style_similarity,
      expected_keyword_similarity,
  ):
    failed_response = generative_models.GenerationResponse.from_dict({
        "candidates": [{
            "finish_reason": generative_models.FinishReason.SAFETY,
        }]
    })

    with testing_utils.PatchGenerativeModel(response=failed_response):
      copycat_instance = copycat.Copycat.create_from_pandas(
          training_data=self.training_data(20),
          embedding_model_name="text-embedding-004",
          ad_format="text_ad",
          vectorstore_exemplar_selection_method="random",
      )

      response = copycat_instance.generate_new_ad_copy(
          keywords=["my keyword 1, my keyword 2"],
          style_guide="This is my style guide.",
          num_in_context_examples=2,
          system_instruction_kwargs=dict(
              company_name="My company",
              language="english",
          ),
          allow_memorised_headlines=False,
          allow_memorised_descriptions=False,
          existing_headlines=[existing_headlines],
          existing_descriptions=[existing_descriptions],
      )[0]

      expected_google_ad = google_ads.GoogleAd(
          headlines=existing_headlines or [],
          descriptions=existing_descriptions or [],
      )

      expected_response = copycat.CopycatResponse(
          google_ad=expected_google_ad,
          keywords="my keyword 1, my keyword 2",
          evaluation_results=copycat.EvaluationResults(
              errors=[str(failed_response.candidates[0])],
              warnings=[],
              headlines_are_memorised=False if existing_headlines else None,
              descriptions_are_memorised=False
              if existing_descriptions
              else None,
              style_similarity=expected_style_similarity,
              keyword_similarity=expected_keyword_similarity,
          ),
      )
      self.assertEqual(response, expected_response)

  def test_to_dict_and_from_dict_returns_same_copycat_instance(self):

    copycat_instance = copycat.Copycat.create_from_pandas(
        training_data=self.training_data(3),
        embedding_model_name="text-embedding-004",
        ad_format="text_ad",
        vectorstore_exemplar_selection_method="random",
    )
    copycat_instance.style_guide = "This is my style guide."

    reloaded_copycat_instance = copycat.Copycat.from_dict(
        copycat_instance.to_dict()
    )

    self.assertTrue(
        testing_utils.copycat_instances_are_equal(
            copycat_instance, reloaded_copycat_instance
        )
    )

  @parameterized.parameters([
      "ad_copy_vectorstore",
      "ad_format",
  ])
  def test_from_dict_raises_key_error_if_required_key_is_missing(
      self, required_key
  ):
    with self.assertRaises(KeyError):
      copycat_instance = copycat.Copycat.create_from_pandas(
          training_data=self.training_data(3),
          embedding_model_name="text-embedding-004",
          ad_format="text_ad",
          vectorstore_exemplar_selection_method="random",
      )
      copycat_instance_dict = copycat_instance.to_dict()
      del copycat_instance_dict[required_key]
      copycat.Copycat.from_dict(copycat_instance_dict)

  def test_to_json_and_from_json_returns_same_copycat_instance(self):

    copycat_instance = copycat.Copycat.create_from_pandas(
        training_data=self.training_data(3),
        embedding_model_name="text-embedding-004",
        ad_format="text_ad",
        vectorstore_exemplar_selection_method="random",
    )
    copycat_instance.style_guide = "This is my style guide."

    reloaded_copycat_instance = copycat.Copycat.from_json(
        copycat_instance.to_json()
    )

    self.assertTrue(
        testing_utils.copycat_instances_are_equal(
            copycat_instance, reloaded_copycat_instance
        )
    )

  @testing_utils.PatchGenerativeModel(
      response=(
          '{"headlines": ["generated headline 1", "generated headline 2"],'
          ' "descriptions": ["generated description"]}'
      )
  )
  def test_generate_new_ad_copy_raises_exception_if_different_number_of_keywords_and_instructions(
      self, generative_model_patcher
  ):
    copycat_instance = copycat.Copycat.create_from_pandas(
        training_data=self.training_data(20),
        embedding_model_name="text-embedding-004",
        ad_format="text_ad",
        vectorstore_exemplar_selection_method="random",
    )
    with self.assertRaisesWithLiteralMatch(
        ValueError,
        "keywords and keywords_specific_instructions must have the same"
        " length.",
    ):
      copycat_instance.generate_new_ad_copy(
          keywords=["my keyword 1, my keyword 2"],
          keywords_specific_instructions=[
              "Some keyword specific instructions.",
              "another set",
          ],
          style_guide="This is my style guide.",
          num_in_context_examples=2,
          system_instruction_kwargs=dict(
              company_name="My company",
              language="english",
          ),
      )

  @testing_utils.PatchGenerativeModel(
      response=(
          '{"headlines": ["generated headline 1", "generated headline 2"],'
          ' "descriptions": ["generated description"]}'
      )
  )
  def test_generate_new_ad_copy_raises_exception_if_different_number_of_keywords_and_existing_headlines(
      self, generative_model_patcher
  ):
    copycat_instance = copycat.Copycat.create_from_pandas(
        training_data=self.training_data(20),
        embedding_model_name="text-embedding-004",
        ad_format="text_ad",
        vectorstore_exemplar_selection_method="random",
    )
    with self.assertRaisesWithLiteralMatch(
        ValueError,
        "keywords and existing_headlines must have the same length.",
    ):
      copycat_instance.generate_new_ad_copy(
          keywords=["my keyword 1, my keyword 2"],
          existing_headlines=[
              ["headline 1"],
              ["headline 2"],
          ],
          style_guide="This is my style guide.",
          num_in_context_examples=2,
          system_instruction_kwargs=dict(
              company_name="My company",
              language="english",
          ),
      )

  @testing_utils.PatchGenerativeModel(
      response=(
          '{"headlines": ["generated headline 1", "generated headline 2"],'
          ' "descriptions": ["generated description"]}'
      )
  )
  def test_generate_new_ad_copy_raises_exception_if_different_number_of_keywords_and_existing_descriptions(
      self, generative_model_patcher
  ):
    copycat_instance = copycat.Copycat.create_from_pandas(
        training_data=self.training_data(20),
        embedding_model_name="text-embedding-004",
        ad_format="text_ad",
        vectorstore_exemplar_selection_method="random",
    )
    with self.assertRaisesWithLiteralMatch(
        ValueError,
        "keywords and existing_descriptions must have the same length.",
    ):
      copycat_instance.generate_new_ad_copy(
          keywords=["my keyword 1, my keyword 2"],
          existing_descriptions=[
              ["description 1"],
              ["description 2"],
          ],
          style_guide="This is my style guide.",
          num_in_context_examples=2,
          system_instruction_kwargs=dict(
              company_name="My company",
              language="english",
          ),
      )

  @parameterized.named_parameters([
      dict(
          testcase_name="with_keyword_specific_instructions",
          keywords_specific_instructions=[
              "Some keyword specific instructions."
          ],
          existing_headlines=None,
          existing_descriptions=None,
          expected_final_message=(
              "For the next set of keywords, please consider the following"
              " additional instructions:\n\nSome keyword specific"
              " instructions.\n\nPlease write 3 headlines and 2 descriptions"
              " for this ad.\n\nKeywords: my keyword 1, my keyword 2"
          ),
      ),
      dict(
          testcase_name="without_keyword_specific_instructions",
          keywords_specific_instructions=None,
          existing_headlines=None,
          existing_descriptions=None,
          expected_final_message=(
              "Please write 3 headlines and 2 descriptions"
              " for this ad.\n\nKeywords: my keyword 1, my keyword 2"
          ),
      ),
      dict(
          testcase_name="with_existing_headlines_and_descriptions",
          keywords_specific_instructions=None,
          existing_headlines=[["existing headline 1"]],
          existing_descriptions=[
              ["existing description 1"],
          ],
          expected_final_message=(
              "This ad already has 1 headlines and 1 descriptions:\n\n-"
              " headlines: ['existing headline 1']\n- descriptions: ['existing"
              " description 1']\n\nPlease write 2 more headlines and 1 more"
              " descriptions to complete this ad.\n\nKeywords: my keyword 1, my"
              " keyword 2"
          ),
      ),
  ])
  @testing_utils.PatchGenerativeModel(
      response=(
          '{"headlines": ["generated headline 1", "generated headline 2"],'
          ' "descriptions": ["generated description"]}'
      )
  )
  def test_generate_new_ad_copy_uses_expected_prompt_final_message(
      self,
      generative_model_patcher,
      keywords_specific_instructions,
      existing_headlines,
      existing_descriptions,
      expected_final_message,
  ):
    copycat_instance = copycat.Copycat.create_from_pandas(
        training_data=self.training_data(20),
        embedding_model_name="text-embedding-004",
        ad_format="text_ad",
        vectorstore_exemplar_selection_method="random",
    )

    _ = copycat_instance.generate_new_ad_copy(
        keywords=["my keyword 1, my keyword 2"],
        keywords_specific_instructions=keywords_specific_instructions,
        style_guide="This is my style guide.",
        num_in_context_examples=2,
        system_instruction_kwargs=dict(
            company_name="My company",
            language="english",
        ),
        existing_headlines=existing_headlines,
        existing_descriptions=existing_descriptions,
    )[0]

    mock_generate_content_async = (
        generative_model_patcher.mock_generative_model.generate_content_async
    )

    self.assertEqual(
        mock_generate_content_async.call_args[0][0][-1].parts[0].text,
        expected_final_message,
    )

  @testing_utils.PatchGenerativeModel(
      response=(
          '{"headlines": ["generated headline 1", "generated headline 2"],'
          ' "descriptions": ["generated description"]}'
      )
  )
  def test_generate_new_ad_copy_for_dataframe_returns_expected_response(
      self,
      generative_model_patcher,
  ):
    data = pd.DataFrame({
        "keywords": ["my keyword 1, my keyword 2", "my keyword 3"],
        "keywords_specific_instructions": [
            "Some keyword specific instructions.",
            "",
        ],
        "existing_headlines": [["existing headline 1"], []],
        "existing_descriptions": [["existing description 1"], []],
    })

    copycat_instance = copycat.Copycat.create_from_pandas(
        training_data=self.training_data(20),
        embedding_model_name="text-embedding-004",
        ad_format="text_ad",
        vectorstore_exemplar_selection_method="random",
    )
    response_series = copycat_instance.generate_new_ad_copy_for_dataframe(
        data,
        style_guide="This is my style guide.",
        num_in_context_examples=2,
        system_instruction_kwargs=dict(
            company_name="My company",
            language="english",
        ),
    )

    for response in response_series:
      # I don't want to test the similarity metrics here, so I'm just setting
      # them to None.
      response.evaluation_results.style_similarity = None
      response.evaluation_results.keyword_similarity = None

    expected_response_series = pd.Series([
        copycat.CopycatResponse(
            google_ad=google_ads.GoogleAd(
                headlines=[
                    "existing headline 1",
                    "generated headline 1",
                    "generated headline 2",
                ],
                descriptions=[
                    "existing description 1",
                    "generated description",
                ],
            ),
            keywords="my keyword 1, my keyword 2",
            evaluation_results=copycat.EvaluationResults(
                errors=[],
                warnings=[],
                headlines_are_memorised=False,
                descriptions_are_memorised=False,
                style_similarity=None,
                keyword_similarity=None,
            ),
        ),
        copycat.CopycatResponse(
            google_ad=google_ads.GoogleAd(
                headlines=["generated headline 1", "generated headline 2"],
                descriptions=["generated description"],
            ),
            keywords="my keyword 3",
            evaluation_results=copycat.EvaluationResults(
                errors=[],
                warnings=[],
                headlines_are_memorised=False,
                descriptions_are_memorised=False,
                style_similarity=None,
                keyword_similarity=None,
            ),
        ),
    ])

    pd.testing.assert_series_equal(response_series, expected_response_series)

  @testing_utils.PatchGenerativeModel(
      response=(
          '{"headlines": ["generated headline 1", "generated headline 2"],'
          ' "descriptions": ["generated description"]}'
      )
  )
  def test_generate_new_ad_copy_for_dataframe_returns_expected_response_if_only_keywords_column_is_provided(
      self,
      generative_model_patcher,
  ):
    data = pd.DataFrame({
        "keywords": ["my keyword 1, my keyword 2", "my keyword 3"],
    })

    copycat_instance = copycat.Copycat.create_from_pandas(
        training_data=self.training_data(20),
        embedding_model_name="text-embedding-004",
        ad_format="text_ad",
        vectorstore_exemplar_selection_method="random",
    )
    response_series = copycat_instance.generate_new_ad_copy_for_dataframe(
        data,
        style_guide="This is my style guide.",
        num_in_context_examples=2,
        system_instruction_kwargs=dict(
            company_name="My company",
            language="english",
        ),
    )

    for response in response_series:
      # I don't want to test the similarity metrics here, so I'm just setting
      # them to None.
      response.evaluation_results.style_similarity = None
      response.evaluation_results.keyword_similarity = None

    expected_response_series = pd.Series([
        copycat.CopycatResponse(
            google_ad=google_ads.GoogleAd(
                headlines=[
                    "generated headline 1",
                    "generated headline 2",
                ],
                descriptions=[
                    "generated description",
                ],
            ),
            keywords="my keyword 1, my keyword 2",
            evaluation_results=copycat.EvaluationResults(
                errors=[],
                warnings=[],
                headlines_are_memorised=False,
                descriptions_are_memorised=False,
                style_similarity=None,
                keyword_similarity=None,
            ),
        ),
        copycat.CopycatResponse(
            google_ad=google_ads.GoogleAd(
                headlines=["generated headline 1", "generated headline 2"],
                descriptions=["generated description"],
            ),
            keywords="my keyword 3",
            evaluation_results=copycat.EvaluationResults(
                errors=[],
                warnings=[],
                headlines_are_memorised=False,
                descriptions_are_memorised=False,
                style_similarity=None,
                keyword_similarity=None,
            ),
        ),
    ])

    pd.testing.assert_series_equal(response_series, expected_response_series)

  @testing_utils.PatchGenerativeModel(
      response=(
          '{"headlines": ["generated headline 1", "generated headline 2"],'
          ' "descriptions": ["generated description"]}'
      )
  )
  def test_generate_new_ad_copy_for_dataframe_raises_exception_if_keywords_column_is_missing(
      self,
      generative_model_patcher,
  ):
    data = pd.DataFrame({
        "keywords_specific_instructions": [
            "Some keyword specific instructions.",
            "",
        ],
        "existing_headlines": [["existing headline 1"], []],
        "existing_descriptions": [["existing description 1"], []],
    })

    copycat_instance = copycat.Copycat.create_from_pandas(
        training_data=self.training_data(20),
        embedding_model_name="text-embedding-004",
        ad_format="text_ad",
        vectorstore_exemplar_selection_method="random",
    )
    with self.assertRaisesWithLiteralMatch(
        ValueError,
        "The dataframe does not contain the required column: keywords",
    ):
      copycat_instance.generate_new_ad_copy_for_dataframe(
          data,
          style_guide="This is my style guide.",
          num_in_context_examples=2,
          system_instruction_kwargs=dict(
              company_name="My company",
              language="english",
          ),
      )


if __name__ == "__main__":
  absltest.main()
